import logging
from collections import OrderedDict
from typing import Any, Tuple, Optional, Union

from prefixcommons import curie_util
from rdflib import Namespace, URIRef, Graph, BNode
from rdflib.namespace import is_ncname

META_NS = "meta"
META_URI = "https://w3id.org/biolink/biolinkml/meta"


class Namespaces(OrderedDict):
    """ Namespace manager.  Functions as both a dictionary and a python
     namespace.

     Supports:  namespaces.NS [= uri]
                namespaces[NS] [= uri]
                namespaces._default [= uri]    # namespace for ':'
                namespaces._base [= uri]       # namespace for @base

     Functions: namespaces.curie_for(uri) --> curie
                namespaces.prefix_for(uri_or_curie) --> NCName
                namespaces.add_prefixmap(map name)
     """
    _default_key = '@default'
    _base_key = '@base'

    # BNODE management -- when the namespace is '_', we map the ln via _bnodes to a new bnode
    _bnodes = {}                    # BNode to BNode map
    _empty_bnode = BNode()          # Node for '_:'

    def __init__(self, g: Optional[Graph] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if g is not None:
            self.from_graph(g)

    def __setitem__(self, key, value):
        if key == '_':
            if not value:
                if self._empty_bnode not in self._bnodes:
                    self._bnodes[self._empty_bnode] = BNode()
            elif value in self._bnodes:
                raise ValueError(f"BNode {value} is already mapped to {self[self._bnodes[value]]}")
            else:
                target_bnode = BNode()
                self._bnodes[value] = target_bnode
                super().__setitem__(value, target_bnode)
        elif is_ncname(key):
            v = Namespace(str(value))
            if key in self:
                if self[key] != v:
                    logging.getLogger('Namespaces').\
                        warning(f"{key} namespace is already mapped to {self[key]} - Mapping to {v} ignored")
            else:
                super().__setitem__(key, v)
        else:
            raise ValueError(f"Invalid NCName: {key}")

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key: str, value):
        if key.startswith('_') and len(key) > 1:
            super().__setattr__(key, value)
        else:
            self[key] = value

    @property
    def _default(self) -> Optional[URIRef]:
        return self.get(self._default_key, None)

    @_default.setter
    def _default(self, item: Any) -> None:
        v = Namespace(str(item))
        if self._default is not None:
            if self._default != v:
                raise ValueError(f"Default value is already set to {self._default}")
        else:
            super().__setitem__(self._default_key, v)

    @_default.deleter
    def _default(self) -> None:
        super().__delitem__(self._default_key)

    @property
    def _base(self) -> Optional[URIRef]:
        return self.get(self._base_key, None)

    @_base.setter
    def _base(self, item: Any) -> None:
        v = Namespace(str(item))
        if self._base is not None:
            if self._base != v:
                raise ValueError(f"Base value is already set to {self._base}")
        else:
            super().__setitem__(self._base_key, v)

    @_base.deleter
    def _base(self) -> None:
        super().__delitem__(self._base_key)

    def curie_for(self, uri: Any, default_ok: bool = True, pythonform: bool = False) -> Optional[str]:
        """
        Return the most appropriate CURIE for URI.  The first longest matching prefix used, if any.  If no CURIE is
        present, None is returned

        @param uri: URI to create the CURIE for
        @param default_ok: True means the default prefix is ok. Otherwise we have to have a reql prefix
        @param pythonform: True means take the python/rdflib uppercase format
        """
        if pythonform:
            default_ok = False
        match: Tuple[str, Optional[Namespace]] = ('', None)     # match string / prefix
        u = str(uri)

        # Find the longest match
        for k, v in self.items():
            vs = str(v)
            if u.startswith(vs):
                if len(vs) > len(match[0]) and (default_ok or k not in (Namespaces._default_key, Namespaces._base_key)):
                    match = (vs, k)
        if len(match[0]):
            if pythonform:
                ns = match[1].upper()
                ln = u.replace((match[0]), '')
                if not ln:
                    return f"URIRef(str({ns}))"
                elif ln.isidentifier():
                    return f"{ns}.{ln}"
                else:
                    return f'{ns}["{ln}"]'
            else:
                return u.replace(match[0],
                                 ':' if match[1] == Namespaces._default_key else
                                 '' if match[1] == Namespaces._base_key else match[1] + ':')
        return None

    def prefix_for(self, uri_or_curie: Any) -> Optional[str]:
        return self.prefix_suffix(uri_or_curie)[0]

    def prefix_suffix(self, uri_or_curie: Any) -> Tuple[Optional[str], Optional[str]]:
        uri_or_curie = str(uri_or_curie)
        if ':/' in uri_or_curie:
            uri_or_curie = self.curie_for(uri_or_curie)
        return uri_or_curie.split(':') if uri_or_curie else (None, None)


    def uri_for(self, uri_or_curie: Any) -> URIRef:
        """
        Map a curie or URI into a full URIRef.

        :param uri_or_curie: "NCNAME ':' suffix" or plain URI
        :return: Corresponding URI
        """
        uri_or_curie = str(uri_or_curie)
        if '://' in uri_or_curie:
            return URIRef(uri_or_curie)
        if ':' in uri_or_curie:
            prefix, local = str(uri_or_curie).split(':', 1)
            if not prefix:
                prefix = self._default_key
            elif not is_ncname(prefix):
                raise ValueError(f"Not a valid CURIE: {uri_or_curie}")
        else:
            prefix, local = self._base_key, uri_or_curie

        if prefix not in self:
            raise ValueError(f"Unknown CURIE prefix: {prefix}")
        return URIRef(self.join(self[prefix], local))

    def uri_or_curie_for(self, prefix: Union[str, URIRef], suffix: str) -> str:
        """ Return a CURIE for prefix/suffix in possible, else a URI """
        if isinstance(prefix, URIRef) or ':/' in prefix:
            prefix_as_uri = str(prefix)
            for k, v in self.items():
                if not k.startswith('@') and prefix_as_uri == str(v):
                    return k + ':' + suffix
            return self.join(str(prefix), suffix)
        elif prefix not in self:
            raise ValueError(f"Unrecognized prefix: {prefix}")
        else:
            return prefix + ':' + suffix

    def load_graph(self, g: Graph) -> Graph:
        """ Transfer all of the known namespaces into G """
        for k, v in self.items():
            if not k.startswith('_') and not k.startswith('@'):
                g.bind(k, URIRef(v))
        return g

    def from_graph(self, g: Graph) -> "Namespaces":
        """ Transfer al bindings from G """
        for ns, name in g.namespaces():
            if ns:
                self[ns] = name
            else:
                self._default = name
        return self

    @staticmethod
    def join(prefix: str, suffix: str) -> str:
        return prefix + suffix

    @staticmethod
    def sfx(uri: str) -> str:
        """
        Add a separator to a uri if none exists.

        Note: This should only be used with module id's -- it is not uncommon to use partial prefixes,
        e.g. PREFIX bfo: http://purl.obolibrary.org/obo/BFO_
        :param uri: uri to be suffixed
        :return: URI with suffix
        """
        return str(uri) + ('' if uri.endswith(('/', '#')) else '/')

    def add_prefixmap(self, map_name: str, include_defaults: bool = True) -> None:
        """
        Add a prefixcommons map.  Only prefixes that have not been previously defined are added.

        :param map_name: prefixcommons map name
        :param include_defaults: if True, take defaults from the map.
        :return:
        """
        for k, v in curie_util.read_biocontext(map_name).items():
            if not k:
                if include_defaults and not self._default:
                    self._default = v
            elif k not in self:
                if is_ncname(k):
                    self[k] = v
                else:
                    logging.getLogger('Namespaces').info(f"biocontext map {map_name} has illegal prefix: {k}")
                    pass


def base_namespace(nsm: Namespaces) -> Optional[str]:
    return nsm._base


def default_namespace(nsm: Namespaces) -> Optional[str]:
    return nsm._default



import unittest
from biolinkml.generators.pythongen import PythonGenerator

yaml_txt = """
id: https://issue_test/107/schema
name: schema

default_curi_maps:
    - semweb_context

types:
  string:
    uri: xsd:string
    base: str
    description: A character string

prefixes:
  ICD-9: http://test.org/prefix/ICD9

emit_prefixes:
    - ICD-9
"""


class PrefixTestCase(unittest.TestCase):
    def test_prefix(self):
        python = PythonGenerator(yaml_txt).serialize()
        compile(python, 'test', 'exec')
        self.assertTrue(True, "Passed the generate")


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from biolinkml.generators.markdowngen import MarkdownGenerator
from tests.test_issues import sourcedir, outputdir
from tests.utils.dirutils import make_and_clear_directory


class Issue62TestCase(unittest.TestCase):

    def test_issue_62(self):
        """ Make sure that types are generated as part of the output """
        outdir = os.path.join(outputdir, 'issue62')
        make_and_clear_directory(outdir)
        yaml_fname = os.path.join(sourcedir, 'issue_62.yaml')
        MarkdownGenerator(yaml_fname).serialize(directory=outdir)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from typing import Callable

from jsonasobj import as_json, loads, load, as_dict, JsonObj

from biolinkml.meta import SchemaDefinition
from biolinkml.utils.rawloader import load_raw_schema
from biolinkml.utils.schemaloader import SchemaLoader
from tests.test_utils import inputdir, outputdir


class RawLoaderTestCase(unittest.TestCase):

    def _verify_schema1_content(self, schema: SchemaDefinition, source_file,
                                addl_checks: Callable[[SchemaDefinition], None]=None) -> None:
        expected = loads(f"""{{
           "default_prefix": "http://example.org/{source_file}/",
           "name": "{source_file}",
           "id": "http://example.org/{source_file}",
           "title": "Load Raw Schema Test",
           "metamodel_version": "0.5.0",
           "source_file": "{source_file}.yaml",
           "source_file_date": "Mon Dec 31 11:25:38 2018",
           "source_file_size": 76,
           "generation_date": "2018-12-31 11:50"
        }}""")

        schema.source_file = os.path.basename(schema.source_file)
        if addl_checks:
            addl_checks(schema)
        self.assertTrue(isinstance(schema.metamodel_version, str))
        expected.metamodel_version = schema.metamodel_version
        self.assertTrue(isinstance(schema.source_file_date, str))
        expected.source_file_date = schema.source_file_date
        self.assertTrue(isinstance(schema.source_file_size, int))
        expected.source_file_size = schema.source_file_size
        self.assertTrue(isinstance(schema.generation_date, str))
        expected.generation_date = schema.generation_date
        self.assertEqual(expected, loads(as_json(schema)))

    def test_load_raw_file(self):
        """ Test loading a data file """
        self._verify_schema1_content(load_raw_schema(os.path.join(inputdir, 'schema1.yaml')), 'schema1')

        # Verify that we can't pass source_file parameters when we've got a directory name
        with self.assertRaises(AssertionError):
            load_raw_schema(os.path.join(inputdir, 'schema1.yaml'), source_file_size=117)

    def test_explicit_name(self):
        """ Test the named schema option """
        self._verify_schema1_content(load_raw_schema(os.path.join(inputdir, 'schema2.yaml')), 'schema2')

    def test_multi_schemas(self):
        """ Test multiple schemas in the same file """
        def check_types(s: SchemaDefinition) -> None:
            output = os.path.join(outputdir, 'schema4.json')
            if not os.path.exists(output):
                with open(output, 'w') as f:
                    f.write(as_json(JsonObj(**{k: as_dict(loads(as_json(v))) for k, v in s.types.items()})))
                    self.fail(f"File {output} created - rerun test")

            with open(output) as f:
                expected = as_dict(load(f))
            self.assertEqual(expected, {k: as_dict(loads(as_json(v))) for k, v in s.types.items()})
            s.types = None

        self._verify_schema1_content(load_raw_schema(os.path.join(inputdir, 'schema4.yaml')), 'schema4', check_types)

    def test_base_dir(self):
        """ Test the base directory option  """
        self._verify_schema1_content(load_raw_schema('schema1.yaml', base_dir=inputdir), 'schema1')

    def test_schema_id(self):
        """ Test loading a schema with just an id """
        self._verify_schema1_content(load_raw_schema('schema3.yaml', base_dir=inputdir), 'schema3')

    def test_name_from_sourcefile(self):
        """ Test no identifier at all  """
        with self.assertRaises(ValueError):
            load_raw_schema(os.path.join(inputdir, 'schema5.yaml'))


    def test_load_text(self):
        """ Test loading straight text """
        with open(os.path.join(inputdir, 'schema1.yaml')) as f:
            self._verify_schema1_content(load_raw_schema(f.read(), 'schema1.yaml', "Mon Dec 31 11:25:38 2018", 76),
                                         'schema1')

    def test_representation_errors(self):
        """ Test misformed schema elements """
        fn = os.path.join(inputdir, 'typeerror1.yaml')
        with self.assertRaises(ValueError):
            SchemaLoader(fn)
        fn = os.path.join(inputdir, 'typeerror2.yaml')
        with self.assertRaises(ValueError):
            SchemaLoader(fn)
        fn = os.path.join(inputdir, 'typeerror3.yaml')
        with self.assertRaises(ValueError):
            SchemaLoader(fn)
        fn = os.path.join(inputdir, 'typeerror4.yaml')
        with self.assertRaises(ValueError):
            SchemaLoader(fn)


if __name__ == '__main__':
    unittest.main()



import os
import re
import sys
from typing import Optional, Tuple, List, Union, TextIO, Callable, Dict, Iterator, cast, Set

import click
from rdflib import URIRef

import biolinkml
from biolinkml.generators import PYTHON_GEN_VERSION
from biolinkml.meta import SchemaDefinition, SlotDefinition, ClassDefinition, ClassDefinitionName, \
    SlotDefinitionName, DefinitionName, Element, TypeDefinition, Definition
from biolinkml.utils.formatutils import camelcase, underscore, be, wrapped_annotation, split_line, sfx
from biolinkml.utils.generator import Generator, shared_arguments
from biolinkml.utils.ifabsent_functions import ifabsent_value_declaration, ifabsent_postinit_declaration, \
    default_curie_or_uri
from biolinkml.utils.metamodelcore import builtinnames
from includes import types


class PythonGenerator(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = PYTHON_GEN_VERSION
    valid_formats = ['py']
    visit_all_class_slots = False

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], format: str = valid_formats[0],
                 emit_metadata: bool = True, **kwargs) -> None:
        self.sourcefile = schema
        self.emit_prefixes: Set[str] = set()
        if format is None:
            format = self.valid_formats[0]
        super().__init__(schema, format, emit_metadata=emit_metadata, **kwargs)
        if not self.schema.source_file and isinstance(self.sourcefile, str) and '\n' not in self.sourcefile:
            self.schema.source_file = os.path.basename(self.sourcefile)

    def visit_schema(self, **kwargs) -> None:
        # Add explicitly declared prefixes
        self.emit_prefixes.update([p.prefix_prefix for p in self.schema.prefixes.values()])

        # Add all emit statements
        self.emit_prefixes.update(self.schema.emit_prefixes)

        # Add the default prefix
        if self.schema.default_prefix:
            self.emit_prefixes.add(self.namespaces.prefix_for(self.schema.default_prefix))

    def visit_class(self, cls: ClassDefinition) -> bool:
        if not cls.imported_from:
            cls_prefix = self.namespaces.prefix_for(cls.class_uri)
            if cls_prefix:
                self.emit_prefixes.add(cls_prefix)
            self.add_mappings(cls)
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if not slot.imported_from:
            slot_prefix = self.namespaces.prefix_for(slot.slot_uri)
            if slot_prefix:
                self.emit_prefixes.add(slot_prefix)
            self.add_mappings(slot)

    def visit_type(self, typ: TypeDefinition) -> None:
        if not typ.imported_from:
            type_prefix = self.namespaces.prefix_for(typ.uri)
            if type_prefix:
                self.emit_prefixes.add(type_prefix)

    def add_mappings(self, defn: Definition) -> None:
        """
        Process any mappings in defn, adding all of the mappings prefixes to the namespace map
        :param defn: Class or Slot Definition
        """
        self.add_id_prefixes(defn)
        for mapping in defn.mappings:
            if '://' in mapping:
                mcurie = self.namespaces.curie_for(mapping)
                print(f"No namespace defined for URI: {mapping}", file=sys.stderr)
                if mcurie is None:
                    return        # Absolute path - no prefix/name
                else:
                    mapping = mcurie
            if ':' not in mapping or len(mapping.split(':')) != 2:
                raise ValueError(f"Definition {defn.name} - unrecognized mapping: {mapping}")
            ns = mapping.split(':')[0]
            self.emit_prefixes.add(ns)

    def add_id_prefixes(self, element: Element) -> None:
        self.emit_prefixes.update(element.id_prefixes)

    def gen_schema(self) -> str:
        split_descripton = '\n#              '.join(split_line(be(self.schema.description), split_len=100))
        head = f'''# Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
# Generation date: {self.schema.generation_date}
# Schema: {self.schema.name}
#''' if self.emit_metadata else ''

        return f'''{head}
# id: {self.schema.id}
# description: {split_descripton}
# license: {be(self.schema.license)}

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
{self.gen_imports()}

metamodel_version = "{self.schema.metamodel_version}"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
{self.gen_namespaces()}


# Types
{self.gen_typedefs()}
# Class references
{self.gen_references()}

{self.gen_classdefs()}


# Slots
class slots:
    pass

{self.gen_slots()}'''

    def end_schema(self, **_):
        print(re.sub(r' +\n', '\n', self.gen_schema().replace('\t', '    ')).strip(' '), end='')

    def gen_imports(self) -> str:
        listents = [f"from {k} import {', '.join(v)}" for k, v in self.gen_import_list().items()]
        return '\n'.join(listents)

    def gen_import_list(self) -> Dict[str, List[str]]:
        """
        Generate a list of types to import

        :return: source file followed by elements to import
        """
        class ImportList:
            def __init__(self, schema_location: str):
                self.schema_location = schema_location
                self.v: Dict[str, Set[str]] = {}

            def add_element(self, e: Element) -> None:
                if e.imported_from:
                    if str(e.imported_from) == biolinkml.METATYPE_URI:
                        # TODO: figure out how to make this sort of stuff part of the model
                        self.v.setdefault(types.__name__, set()).add(camelcase(e.name))
                    elif str(e.imported_from) == biolinkml.BIOLINK_MODEL_URI:
                        self.v.setdefault(biolinkml.BIOLINK_MODEL_PYTHON_LOC, set()).add(camelcase(e.name))
                    elif e.imported_from.__contains__('://'):
                        raise ValueError(f"Cannot map {e.imported_from} into a python import statement")
                    else:
                        self.v.setdefault(types.__name__, set()).add(camelcase(e.name))

            def add_entry(self, path: Union[str, URIRef], name: str) -> None:
                path = str(path)
                if path == biolinkml.METATYPE_URI:
                    # TODO: figure out how to make this sort of stuff part of the model
                    self.v.setdefault(types.__name__, set()).add(name)
                elif path == biolinkml.BIOLINK_MODEL_URI:
                    self.v.setdefault(biolinkml.BIOLINK_MODEL_PYTHON_LOC, set()).add(name)
                elif path.__contains__('://'):
                    raise ValueError(f"Cannot map {path} into a python import statement")
                else:
                    self.v.setdefault(path.replace('/', '.'), set()).add(name)

            def values(self) -> Dict[str, List[str]]:
                return {k: sorted(self.v[k]) for k in sorted(self.v.keys())}

        def add_type_ref(typ: TypeDefinition) -> None:
            if not typ.typeof and typ.base and typ.base not in builtinnames:
                if '.' in typ.base:
                    rval.add_entry(*typ.base.rsplit('.'))
                else:
                    rval.add_entry('biolinkml.utils.metamodelcore', typ.base)
            if typ.typeof:
                add_type_ref(self.schema.types[typ.typeof])
            rval.add_element(typ)

        def add_slot_range(slot: SlotDefinition) -> None:
            if slot.range:
                if slot.range in self.schema.types:
                    add_type_ref(self.schema.types[slot.range])
                else:
                    cls = self.schema.classes[slot.range]
                    if cls.imported_from:
                        if self.class_identifier(cls):
                            rval.add_entry(cls.imported_from, self.class_identifier_path(cls, False)[-1])
                        if slot.inlined:
                            rval.add_element(cls)

        rval = ImportList(self.schema_location)
        for typ in self.schema.types.values():
            if not typ.imported_from:
                add_type_ref(typ)
        for slot in self.schema.slots.values():
            if not slot.imported_from:
                if slot.is_a:
                    parent = self.schema.slots[slot.is_a]
                    if parent.imported_from:
                        rval.add_element(self.schema.slots[slot.is_a])
                if slot.domain:
                    domain = self.schema.classes[slot.domain]
                    if domain.imported_from:
                        rval.add_element(self.schema.classes[slot.domain])
                add_slot_range(slot)
        for cls in self.schema.classes.values():
            if not cls.imported_from:
                if cls.is_a:
                    parent = self.schema.classes[cls.is_a]
                    if parent.imported_from:
                        rval.add_element(self.schema.classes[cls.is_a])
                        if self.class_identifier(parent):
                            rval.add_entry(parent.imported_from, self.class_identifier_path(parent, False)[-1])
                for slotname in cls.slots:
                    add_slot_range(self.schema.slots[slotname])

        return rval.values()

    def gen_namespaces(self) -> str:
        dflt_prefix = default_curie_or_uri(self)
        dflt = f"CurieNamespace('', '{sfx(dflt_prefix)}')" if ':/' in dflt_prefix else dflt_prefix.upper()
        return '\n'.join([
            f"{pfx.upper().replace('.', '_').replace('-', '_')} = CurieNamespace('{pfx.replace('.', '_')}', '{self.namespaces[pfx]}')"
            for pfx in sorted(self.emit_prefixes) if pfx in self.namespaces
        ] + [f"DEFAULT_ = {dflt}"])


    def gen_references(self) -> str:
        """ Generate python type declarations for all identifiers (primary keys)
        """
        rval = []
        for cls in self._sort_classes(self.schema.classes.values()):
            if not cls.imported_from:
                pkeys = self.primary_keys_for(cls)
                if pkeys:
                    for pk in pkeys:
                        classname = camelcase(cls.name) + camelcase(self.aliased_slot_name(pk))
                        if cls.is_a and self.class_identifier(cls.is_a):
                            parents = self.class_identifier_path(cls.is_a, False)
                        else:
                            parents = self.slot_type_path(self.schema.slots[pk])
                        # TODO: check if parents[-1] is str, float, or int. changed it to extedned_
                        parent_cls = 'extended_' + parents[-1] if parents[-1] in ['str', 'float', 'int'] else parents[-1]
                        rval.append(f'class {classname}({parent_cls}):\n\tpass')
                        break       # We only do the first primary key
        return '\n\n\n'.join(rval)

    def gen_typedefs(self) -> str:
        """ Generate python type declarations for all defined types """
        rval = []
        for typ in self.schema.types.values():
            if not typ.imported_from:
                typname = camelcase(typ.name)
                desc = f'\n\t""" {typ.description} """' if typ.description else ''

                if typ.typeof:
                    parent_typename = camelcase(typ.typeof)
                    rval.append(f'class {typname}({parent_typename}):{desc}\n\t{self.gen_type_meta(typ)}\n\n')
                else:
                    base_base = typ.base.rsplit('.')[-1]
                    rval.append(f'class {typname}({base_base}):{desc}\n\t{self.gen_type_meta(typ)}\n\n')
        return '\n'.join(rval)

    def gen_classdefs(self) -> str:
        """ Create class definitions for all non-mixin classes in the model
            Note that apply_to classes are transformed to mixins
        """
        clist = self._sort_classes(self.schema.classes.values())
        return '\n'.join([self.gen_classdef(v) for v in clist
                          if not v.mixin and not v.imported_from])

    def gen_classdef(self, cls: ClassDefinition) -> str:
        """ Generate python definition for class cls """

        parentref = f'({self.formatted_element_name(cls.is_a, True) if cls.is_a else "YAMLRoot"})'
        slotdefs = self.gen_class_variables(cls)
        postinits = self.gen_postinits(cls)

        wrapped_description = f'\n\t"""\n\t{wrapped_annotation(be(cls.description))}\n\t"""' if be(cls.description) else ''

        return ('\n@dataclass' if slotdefs else '') + \
               f'\nclass {self.class_or_type_name(cls.name)}{parentref}:{wrapped_description}' + \
               f'\n\t{self.gen_inherited_slots(cls)}\n' + \
               f'\n\t{self.gen_class_meta(cls)}\n' + \
               (f'\n\t{slotdefs}' if slotdefs else '') + \
               (f'\n{postinits}' if postinits else '')

    def gen_inherited_slots(self, cls: ClassDefinition) -> str:
        inherited_slots = []
        for slotname in cls.slots:
            slot = self.schema.slots[slotname]
            if slot.inherited:
                inherited_slots.append(slot.alias if slot.alias else slotname)
        inherited_slots_str = ", ".join([f'"{underscore(s)}"' for s in inherited_slots])
        return f"_inherited_slots: ClassVar[List[str]] = [{inherited_slots_str}]"

    def gen_class_meta(self, cls: ClassDefinition) -> str:
        class_class_uri = self.namespaces.uri_for(cls.class_uri)
        if class_class_uri:
            cls_python_uri = self.namespaces.curie_for(class_class_uri, default_ok=False, pythonform=True)
            class_class_curie = self.namespaces.curie_for(class_class_uri, default_ok=False, pythonform=False)
        else:
            cls_python_uri = None
            class_class_curie = None
        if class_class_curie:
            class_class_curie = f'"{class_class_curie}"'
        class_class_uri = cls_python_uri if cls_python_uri else f'URIRef("{class_class_uri}")'
        class_model_uri = self.namespaces.uri_or_curie_for(self.schema.default_prefix or "DEFAULT_", camelcase(cls.name))
        if ':/' in class_model_uri:
            class_model_uri = f'URIRef("{class_model_uri}")'
        else:
            ns, ln = class_model_uri.split(':', 1)
            class_model_uri = f"{ns.upper()}.{ln}"

        vars = [f'class_class_uri: ClassVar[URIRef] = {class_class_uri}',
                f'class_class_curie: ClassVar[str] = {class_class_curie}',
                f'class_name: ClassVar[str] = "{cls.name}"',
                f'class_model_uri: ClassVar[URIRef] = {class_model_uri}']
        return "\n\t".join(vars)

    def gen_type_meta(self, typ: TypeDefinition) -> str:
        type_class_uri = self.namespaces.uri_for(typ.uri)
        if type_class_uri:
            type_python_uri = self.namespaces.curie_for(type_class_uri, default_ok=False, pythonform=True)
            type_class_curie = self.namespaces.curie_for(type_class_uri, default_ok=False, pythonform=False)
        else:
            type_python_uri = None
            type_class_curie = None
        if type_class_curie:
            type_class_curie = f'"{type_class_curie}"'
        type_class_uri = type_python_uri if type_python_uri else f'URIRef("{type_class_uri}")'
        type_model_uri = self.namespaces.uri_or_curie_for(self.schema.default_prefix, camelcase(typ.name))
        if ':/' in type_model_uri:
            type_model_uri = f'URIRef("{type_model_uri}")'
        else:
            ns, ln = type_model_uri.split(':', 1)
            ln_suffix = f".{ln}" if ln.isidentifier() else f'["{ln}"]'
            type_model_uri = f"{ns.upper()}{ln_suffix}"
        vars = [f'type_class_uri = {type_class_uri}',
                f'type_class_curie = {type_class_curie}',
                f'type_name = "{typ.name}"',
                f'type_model_uri = {type_model_uri}']
        return "\n\t".join(vars)


    def gen_class_variables(self,
                            cls: ClassDefinition) -> str:
        """
        Generate the variable declarations for a dataclass.

        :param cls: class containing variables to be rendered in inheritence hierarchy
        :return: variable declarations for target class and its ancestors
        """
        def overridden_slot(slotname: SlotDefinitionName) -> bool:
            """ Determine whether slotname is overridden in any of the descendant classes """
            return False
            # return bool(set(ancestor_path)
            #             .intersection(set(self.synopsis.slotusages.get(self.aliased_slot_name(slotname), []))))

        initializers = []

        is_root = not cls.is_a
        domain_slots = self.domain_slots(cls)

        # Root keys and identifiers go first.  Note that even if a key or identifier is overridden it still
        # appears at the top of the list, as we need to keep the position
        slot_variables = self._slot_iter(cls, lambda slot: (slot.identifier or slot.key) and not slot.ifabsent,
                                         first_hit_only=True)
        initializers += [self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables]

        # Required slots
        slot_variables = self._slot_iter(cls,
                                         lambda slot: slot.required and not slot.identifier and not slot.key and not slot.ifabsent)
        initializers += [self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables]

        # Required or key slots with default values
        slot_variables = self._slot_iter(cls,
                                         lambda slot: slot.ifabsent and slot.required)
        initializers += [self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables]

        # Followed by everything else
        slot_variables = self._slot_iter(cls, lambda slot: not slot.required and not overridden_slot(slot.name)
                                                           and slot in domain_slots)
        initializers += [self.gen_class_variable(cls, slot, True) for slot in slot_variables]

        return '\n\t'.join(initializers)

    def gen_class_variable(self, cls: ClassDefinition, slot: SlotDefinition, can_be_positional: bool) -> str:
        """
        Generate a class variable declaration for the supplied slot

        :param cls: Owning class
        :param slot: slot definition
        :param can_be_positional: True means that positional parameters are allowed
        :return: Initializer string
        """
        slotname = self.slot_name(slot.name)
        slot_range, default_val = self.range_cardinality(slot, cls, can_be_positional)
        ifabsent_text = ifabsent_value_declaration(slot.ifabsent, self, cls, slot) if slot.ifabsent is not None else None
        if ifabsent_text:
            default = f'= {ifabsent_text}'
        else:
            default = f'= {default_val}' if default_val else ''
        return f'''{slotname}: {slot_range} {default}'''

    def range_cardinality(self, slot: SlotDefinition, cls: Optional[ClassDefinition], positional_allowed: bool) \
            -> Tuple[str, Optional[str]]:
        """
        Return the range type including initializers, etc.

        :param slot: slot to generate type for
        :param cls: containing class -- used to render key slots correctly.  If absent, slot is an add-in
        :param positional_allowed: True Means that we are in the positional space
        :return: python property name and initializer (if any)
        """
        range_type, parent_type, _ = self.class_reference_type(slot, cls)

        if slot.multivalued:
            pkey = self.class_identifier(slot.range, keys_count=True)
            if self.is_key_value_class(cast(DefinitionName, slot.range)):
                return range_type, 'empty_dict()'
            elif slot.inlined and pkey:
                base_key = self.gen_class_reference(self.class_identifier_path(slot.range, False))
                return f'Dict[{base_key}, {range_type}]', 'empty_dict()'
            else:
                return f'List[{range_type}]', 'empty_list()'
        elif slot.required:
            return range_type, ('None' if positional_allowed else None)
        else:
            return f'Optional[{range_type}]', 'None'

    def class_reference_type(self, slot: SlotDefinition, cls: Optional[ClassDefinition]) \
            -> Tuple[str, str, str]:
        """
        Return the type of a slot referencing a class

        :param slot: slot to be typed
        :param cls: owning class.  Used for generating key references
        :return: Python class reference type, most proximal type, most proximal type name
        """
        assert not (slot.key or slot.identifier) or cls, "Key slots must have a domain"
        rangelist = self.class_identifier_path(cls, False) if slot.key or slot.identifier else self.slot_type_path(slot)
        prox_type = self.slot_type_path(slot)[-1].rsplit('.')[-1]
        prox_type_name = rangelist[-1]

        # Python version < 3.7 requires quoting forward references
        if cls and slot.inlined and slot.range in self.schema.classes and self.forward_reference(slot.range, cls.name):
            rangelist[-1] = f'"{rangelist[-1]}"'
        return f"{self.gen_class_reference(rangelist)}", prox_type, prox_type_name

    @staticmethod
    def gen_class_reference(rangelist: List[str]) -> str:
        """
        Return a basic or a union type depending on the number of elements in range list

        :param rangelist: List of types from distal to proximal
        :return:
        """
        base = rangelist[0].rsplit('.')[-1]
        return f"Union[{base}, {rangelist[-1]}]" if len(rangelist) > 1 else base

    def gen_postinits(self, cls: ClassDefinition) -> str:
        """ Generate all the typing and existence checks post initialize
        """
        post_inits_pre_super = []
        for slot in self.domain_slots(cls):
            if slot.ifabsent:
                dflt = ifabsent_postinit_declaration(slot.ifabsent, self, cls, slot)
                if dflt and dflt != "None":
                    post_inits_pre_super.append(f'if self.{self.slot_name(slot.name)} is None:')
                    post_inits_pre_super.append(f'\tself.{self.slot_name(slot.name)} = {dflt}')

        post_inits = []
        if not cls.abstract:
            pkeys = self.primary_keys_for(cls)
            for pkey in pkeys:
                slot = self.schema.slots[pkey]
                # TODO: Remove the bypass whenever we get default_range fixed
                if not slot.ifabsent or True:
                    post_inits.append(self.gen_postinit(cls, slot))
        else:
            pkeys = []
        for slot in self.domain_slots(cls):
            # TODO: Remove the bypass whenever we get default_range fixed
            if slot.name not in pkeys and (not slot.ifabsent or True):
                post_inits.append(self.gen_postinit(cls, slot))

        post_inits_pre_super_line = '\n\t\t'.join([p for p in post_inits_pre_super if p]) + \
                                    ('\n\t\t' if post_inits_pre_super else '')
        post_inits_line = '\n\t\t'.join([p for p in post_inits if p])
        return (f'''
    def __post_init__(self, **kwargs: Dict[str, Any]):
        {post_inits_pre_super_line}{post_inits_line}
        super().__post_init__(**kwargs)''' + '\n') if post_inits_line or post_inits_pre_super_line else ''

    # sort classes such that if C is a child of P then C appears after P in the list
    def _sort_classes(self, clist: List[ClassDefinition]) -> List[ClassDefinition]:
        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a is None:
                    can_add = True
                else:
                    if candidate.is_a in [p.name for p in slist]:
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise (f'could not find suitable element in {clist} that does not ref {slist}')
        return slist

    def is_key_value_class(self, range_name: DefinitionName) -> bool:
        """
        Return True if range_name references a class with exactly one key and one value

        :param range_name: class definition (name)
        :return: True if meets the special case
        """
        rng = self.schema.classes.get(range_name)
        if rng:
            pkeys = self.primary_keys_for(rng)
            if pkeys:
                return len(rng.slots) - len(pkeys) == 1
        return False

    def gen_postinit(self, cls: ClassDefinition, slot: SlotDefinition) -> Optional[str]:
        """ Generate python post init rules for slot in class
        """
        rlines: List[str] = []
        slotname = self.slot_name(slot.name)
        range_type_name, base_type, base_type_name = self.class_reference_type(slot, cls)
        single_typed = range_type_name == base_type
        root_definition = not cls.is_a

        # Generate existence check for required slots.  Note that inherited classes have to check post-init because
        # named variables can't be mixed in the class signature
        if slot.required:
            # If we have a root class, the required part is set by the type.  If it is inherited, we have to add
            # the following
            if not slot.multivalued:
                rlines.append(f'if self.{slotname} is None:')
                rlines.append(f'\traise ValueError(f"{slotname} must be supplied")')
            else:
                rlines.append(f'if not isinstance(self.{slotname}, list) or len(self.{slotname}) == 0:')
                rlines.append(f'\traise ValueError(f"{slotname} must be a non-empty list")')
        if slot.range in self.schema.classes or slot.range in self.schema.types:
            indent = len(f'self.{slotname} = [') * ' '
            if not slot.multivalued:
                if not single_typed:
                    if slot.required:
                        rlines.append(f'if not isinstance(self.{slotname}, {base_type_name}):')
                    else:
                        rlines.append(f'if self.{slotname} is not None and '
                                      f'not isinstance(self.{slotname}, {base_type_name}):')
                    # Another really wierd case -- a class that has no properties
                    if slot.range in self.schema.classes and not self.schema.classes[slot.range].slots:
                        rlines.append(f'\tself.{slotname} = {base_type_name}()')
                    else:
                        if self.class_identifier(slot.range) or slot.range in self.schema.types:
                            rlines.append(f'\tself.{slotname} = {base_type_name}(self.{slotname})')
                        else:
                            rlines.append(f'\tself.{slotname} = {base_type_name}(**self.{slotname})')
            elif slot.inlined:
                slot_range_cls = self.schema.classes[slot.range]
                pkeys = self.primary_keys_for(slot_range_cls)
                if pkeys:
                    # Special situation -- if there are only two values: primary key and value,
                    # we load it is a list, not a dictionary
                    if self.is_key_value_class(cast(DefinitionName, slot.range)):
                        class_init = '(k, v)'
                    else:
                        pkey_name = self.formatted_element_name(pkeys[0])
                        class_init = f'({pkey_name}=k, **({{}} if v is None else v))'
                    rlines.append(f'for k, v in self.{slotname}.items():')
                    rlines.append(f'\tif not isinstance(v, {base_type_name}):')
                    rlines.append(f'\t\tself.{slotname}[k] = {base_type_name}{class_init}')
                elif not single_typed:
                    rlines.append(f'self.{slotname} = [v if isinstance(v, {base_type_name})')
                    rlines.append(f'{indent}else {base_type_name}(**v) for v in self.{slotname}]')
            elif not single_typed:
                rlines.append(f'self.{slotname} = [v if isinstance(v, {base_type_name})')
                rlines.append(f'{indent}else {base_type_name}(v) for v in self.{slotname}]')
        return '\n\t\t'.join(rlines)


    def _slot_iter(self, cls: ClassDefinition, test: Callable[[SlotDefinition], bool], first_hit_only: bool = False) \
            -> Iterator[SlotDefinition]:
        """ Return the representation for the set of own slots in cls that pass test

        :param cls: Class containing a set of slots
        :param test: Slot test function
        :param first_hit_only: True means stop on first match.  False means generate all
        :return: Set of slots that match
        """
        for slot in self.all_slots(cls):
            if test(slot):
                yield slot
                if first_hit_only:
                    break

    def primary_keys_for(self, cls: ClassDefinition) -> List[SlotDefinitionName]:
        """ Return the primary key for cls

        @param cls: class to get keys for
        @return: List of primary keys
        """
        return [slot_name for slot_name in cls.slots
                if self.schema.slots[slot_name].key or self.schema.slots[slot_name].identifier]

    def key_name_for(self, class_name: ClassDefinitionName) -> Optional[str]:
        for slot_name in self.primary_keys_for(self.schema.classes[class_name]):
            return self.formatted_element_name(class_name, True) + camelcase(slot_name)
        return None

    def range_type_name(self, slot: SlotDefinition) -> str:
        """ Generate the type name for the slot """
        cidpath = self.slot_type_path(slot)
        if len(cidpath) < 2:
            return cidpath[0]
        else:
            return f"Union[{cidpath[0]}, {cidpath[-1]}]"

    def forward_reference(self, slot_range: str, owning_class: str) -> bool:
        """ Determine whether slot_range is a forward reference """
        if slot_range in self.schema.classes and self.schema.classes[slot_range].imported_from:
            return False
        for cname in self.schema.classes:
            if cname == owning_class:
                return True             # Occurs on or after
            elif cname == slot_range:
                return False            # Occurs before
        return True

    def python_uri_for(self, uriorcurie: Union[str, URIRef]) -> Tuple[str, Optional[str]]:
        """ Return the python form of uriorcurie
        :param uriorcurie:
        :return: URI and CURIE form
        """
        ns, ln = self.namespaces.prefix_suffix(uriorcurie)
        if ns == '':
            ns = 'DEFAULT_'
        if ns is None:
            return f'"str(uriorcurie)"', None
        return ns.upper() + (f".{ln}" if ln.isidentifier() else f"['{ln}']"), ns.upper() + f".curie('{ln}')"


    def gen_slots(self) -> str:
        return '\n\n'.join([self.gen_slot(slot) for slot in self.schema.slots.values() if not slot.imported_from])

    def gen_slot(self, slot: SlotDefinition) -> str:
        python_slot_name = underscore(slot.name)
        slot_uri, slot_curie = self.python_uri_for(slot.slot_uri)
        slot_model_uri, slot_model_curie = \
            self.python_uri_for(self.namespaces.uri_or_curie_for(self.schema.default_prefix, python_slot_name))
        domain = camelcase(slot.domain) if slot.domain and not self.schema.classes[slot.domain].mixin else "None"
        # Going to omit the range on keys where the domain isn't specified (for now)
        if slot.domain is None and (slot.key or slot.identifier):
            rnge = "URIRef"
        else:
            rnge, _ = self.range_cardinality(slot, self.schema.classes[slot.domain] if slot.domain else None, True)
        if slot.mappings:
            map_texts = [self.namespaces.curie_for(self.namespaces.uri_for(m), default_ok=True, pythonform=True)
                         for m in slot.mappings if m != slot.slot_uri]
        else:
            map_texts = []
        if map_texts:
            mappings = ', mappings = [' + ', '.join(map_texts)+ ']'
        else:
            mappings = ''
        return f"""slots.{python_slot_name} = Slot(uri={slot_uri}, name="{slot.name}", curie={slot_curie},
                      model_uri={slot_model_uri}, domain={domain}, range={rnge}{mappings})"""


@shared_arguments(PythonGenerator)
@click.command()
@click.option("--head/--no-head", default=True, help="Emit metadata heading")
def cli(yamlfile, head=True, **args):
    """ Generate python classes to represent a biolink model """
    print(PythonGenerator(yamlfile, emit_metadata=head, **args).serialize(emit_metadata=head, **args))



import os
import unittest
from types import ModuleType
from typing import List

from pyshex import ShExEvaluator, PrefixLibrary
from pyshex.shex_evaluator import EvaluationResult
from rdflib import Graph, Namespace

from biolinkml import METAMODEL_NAMESPACE, LOCAL_SHEXJ_FILE_NAME, LOCAL_METAMODEL_LDCONTEXT_FILE
from biolinkml.generators.jsonldcontextgen import ContextGenerator
from biolinkml.generators.jsonldgen import JSONLDGenerator
from biolinkml.generators.markdowngen import MarkdownGenerator
from biolinkml.generators.pythongen import PythonGenerator
from biolinkml.generators.rdfgen import RDFGenerator
from tests import targetdir, sourcedir, DO_SHEX_VALIDATION
from tests.utils.generator_utils import GeneratorTestCase, BIOLINK_IMPORT_MAP
from tests.utils.metadata_filters import metadata_filter, ldcontext_metadata_filter, json_metadata_filter, \
    json_metadata_context_filter


class MappingsGeneratorTestCase(GeneratorTestCase):
    """ Validate that mappings are available on both a meta and a model level basis """

    source_path = sourcedir
    target_path = targetdir
    model_path = sourcedir
    model_name = 'meta_mappings'
    importmap = BIOLINK_IMPORT_MAP

    def test_mappings_in_metamodel(self):
        """ Make sure that mappings can be specified in a metamodel """
        self.single_file_generator('py', PythonGenerator, filtr=metadata_filter)

        # Make sure the python is valid
        with open(os.path.join(self.source_path, 'meta_mappings.py')) as f:
            pydata = f.read()
        spec = compile(pydata, 'test', 'exec')
        module = ModuleType('test')
        exec(spec, module.__dict__)

    def test_mappings_markdown(self):
        """ Test mappings markdown for  """
        self.directory_generator('meta_mappings_docs', MarkdownGenerator)

    @staticmethod
    def _evaluate_shex_results(results: List[EvaluationResult]) -> bool:
        """
        Check the results of the ShEx evaluation
        :param results: evaluate output
        :return: success indicator
        """
        success = all(r.result for r in results)
        if not success:
            for r in results:
                if not r.result:
                    print(r.reason)
        return success

    @unittest.skipIf(False, "We still need to figure out what to do here")
    def test_mappings_rdf(self):
        """ Test the imported mappings in the biolink metamodel """

        # Generate context and use it to create the RDF
        self.single_file_generator('jsonld', ContextGenerator, filtr=ldcontext_metadata_filter)

        # Generate a copy of the JSON representation of the model
        context_loc = os.path.join(self.source_path, self.model_name + ".jsonld")
        context_args = {"context": ['file://' + LOCAL_METAMODEL_LDCONTEXT_FILE, 'file://' + context_loc]}
        self.single_file_generator('json', JSONLDGenerator,  serialize_args=context_args,  filtr=json_metadata_context_filter)

        # Make a fresh copy of the RDF and validate it as well
        self.single_file_generator('ttl', RDFGenerator, serialize_args=context_args,
                                   comparator=GeneratorTestCase.rdf_comparator)

        g = Graph()
        rdf_file = os.path.join(sourcedir, 'meta_mappings.ttl')
        g.load(rdf_file, format='turtle')
        ns = PrefixLibrary()
        ns.add_rdf(g)
        ns['FULL'] = "http://example.org/fulluri/"
        ns['EX'] = "http://example.org/mappings/"
        ns['META'] = "https://w3id.org/biolink/biolinkml/meta/"
        # Make sure that the expected triples got added

        self.assertEqual({ns.EX.slot1_close, ns.FULL.slot1_close}, set(g.objects(ns.EX.s1, ns.SKOS.closeMatch)))
        self.assertEqual({ns.EX.slot1, ns.FULL.slot1}, set(g.objects(ns.EX.s1, ns.SKOS.exactMatch)))
        self.assertEqual(ns.EX.s3, g.value(ns.EX.s1, ns.META.deprecated_element_has_exact_replacement, any=False))
        self.assertEqual(ns.EX.s4, g.value(ns.EX.s1, ns.META.deprecated_element_has_possible_replacement, any=False))

        self.assertEqual({ns.EX.class1_close, ns.FULL.class1_close}, set(g.objects(ns.EX.C1, ns.SKOS.closeMatch)))
        self.assertEqual({ns.EX.class1, ns.FULL.class1}, set(g.objects(ns.EX.C1, ns.SKOS.exactMatch)))
        self.assertEqual(ns.EX.c2, g.value(ns.EX.C1, ns.META.deprecated_element_has_exact_replacement, any=False))
        self.assertEqual(ns.EX.c3, g.value(ns.EX.C1, ns.META.deprecated_element_has_possible_replacement, any=False))
        if DO_SHEX_VALIDATION:
            EX = Namespace("http://example.org/mappings/")
            focus = EX.testMetamodelMappings
            start = METAMODEL_NAMESPACE.SchemaDefinition
            results = ShExEvaluator(g, LOCAL_SHEXJ_FILE_NAME, focus, start).evaluate(debug=False)
            self.assertTrue(self._evaluate_shex_results(results))
        else:
            print("*** RDF Model validation step was skipped. Set: tests.__init__.DO_SHEX_VALIDATION to run it")


if __name__ == '__main__':
    unittest.main()



# Auto generated from mappings.yaml by pythongen.py version: 0.4.0
# Generation date: 2020-01-27 13:57
# Schema: mappings
#
# id: https://w3id.org/biolink/biolinkml/mappings
# description: Biolink model for mappings
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
from biolinkml.utils.metamodelcore import URIorCURIE
from includes.types import Uriorcurie

metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
IAO = CurieNamespace('IAO', 'http://purl.obolibrary.org/obo/IAO_')
OIO = CurieNamespace('OIO', 'http://www.geneontology.org/formats/oboInOwl#')
BIOLINKML = CurieNamespace('biolinkml', 'https://w3id.org/biolink/biolinkml/')
META = CurieNamespace('meta', 'https://w3id.org/biolink/biolinkml/meta/')
METATYPE = CurieNamespace('metatype', 'https://w3id.org/biolink/biolinkml/type/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = META


# Types

# Class references





# Slots
class slots:
    pass

slots.mappings = Slot(uri=SKOS.mappingRelation, name="mappings", curie=SKOS.curie('mappingRelation'),
                      model_uri=META.mappings, domain=None, range=List[Union[str, URIorCURIE]])

slots.exact_mappings = Slot(uri=SKOS.exactMatch, name="exact mappings", curie=SKOS.curie('exactMatch'),
                      model_uri=META.exact_mappings, domain=None, range=List[Union[str, URIorCURIE]])

slots.close_mappings = Slot(uri=SKOS.closeMatch, name="close mappings", curie=SKOS.curie('closeMatch'),
                      model_uri=META.close_mappings, domain=None, range=List[Union[str, URIorCURIE]])

slots.related_mappings = Slot(uri=SKOS.relatedMatch, name="related mappings", curie=SKOS.curie('relatedMatch'),
                      model_uri=META.related_mappings, domain=None, range=List[Union[str, URIorCURIE]])

slots.deprecated_element_has_exact_replacement = Slot(uri=META.deprecated_element_has_exact_replacement, name="deprecated element has exact replacement", curie=META.curie('deprecated_element_has_exact_replacement'),
                      model_uri=META.deprecated_element_has_exact_replacement, domain=None, range=Optional[Union[str, URIorCURIE]], mappings = [IAO["0100001"]])

slots.deprecated_element_has_possible_replacement = Slot(uri=META.deprecated_element_has_possible_replacement, name="deprecated element has possible replacement", curie=META.curie('deprecated_element_has_possible_replacement'),
                      model_uri=META.deprecated_element_has_possible_replacement, domain=None, range=Optional[Union[str, URIorCURIE]], mappings = [OIO.consider])


import os
import unittest

import yaml
from jsonasobj import as_json, loads, load

from biolinkml.utils.rawloader import load_raw_schema
from biolinkml.utils.yamlutils import DupCheckYamlLoader, as_yaml
from tests.test_utils import inputdir, outputdir
from tests.test_utils.support.base import Base


class YamlUtilTestCase(Base):

    def test_dupcheck_loader(self):
        """ Make sure the duplicate checker finds duplicates """
        with open(os.path.join(inputdir, 'yaml1.yaml')) as f:
            y1 = yaml.safe_load(f)
            self.assertEqual(17, y1['f1'])
        with open(os.path.join(inputdir, 'yaml1.yaml')) as f:
            with self.assertRaises(ValueError):
                yaml.load(f, DupCheckYamlLoader)
        with open(os.path.join(inputdir, 'yaml2.yaml')) as f:
            with self.assertRaises(ValueError):
                yaml.load(f, DupCheckYamlLoader)
        with open(os.path.join(inputdir, 'schema1.yaml')) as f:
            s1 = yaml.load(f, DupCheckYamlLoader)
            self.assertEqual('schema1', s1['name'])

    def test_as_json(self):
        schema = self.fix_schema_metadata(load_raw_schema(os.path.join(inputdir, 'schema6.yaml')))
        outfile = os.path.join(outputdir, 'schema6.json')
        if not os.path.exists(outfile):
            with open(outfile, 'w') as f:
                f.write(as_json(schema))
                self.fail(f"Generated {outfile} - run test again")
        else:
            self.assertEqual(load(outfile), loads(as_json(schema)))

    @unittest.skipIf(True, "Need to make yaml dump work again")
    def test_as_yaml(self):
        """ Test the YAML output representation """
        schema = self.fix_schema_metadata(load_raw_schema(os.path.join(inputdir, 'schema4.yaml')))
        outfile = os.path.join(outputdir, 'schema4.yaml')
        if not os.path.exists(outfile):
            with open(outfile, 'w') as f:
                f.write(as_yaml(schema))
            self.fail(f"Output file: {outfile} created - run test again")

        with open(outfile) as f:
            self.assertEqual(f.read(), as_yaml(schema))


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from biolinkml.generators.pythongen import PythonGenerator
from biolinkml.utils.yamlutils import as_yaml
from tests.test_issues import sourcedir
import yaml


class IssueYamlSerializerTestCase(unittest.TestCase):
    def test_roundtrip(self):
        yaml_fname = os.path.join(sourcedir, 'issue_134.yaml')

        gen = PythonGenerator(yaml_fname)
        schema = gen.schema
        yaml_str = as_yaml(schema)
        generated = yaml.safe_load(yaml_str)
        with open(yaml_fname) as yaml_file:
            original = yaml.safe_load(yaml_file)

        # The generated YAML contains many added fields. Some with default values. Therefore, we can't directly
        # compare it to the original.
        for key in original:
            self.assertEqual(len(original[key]), len(generated[key]))
            if isinstance(original[key], dict):
                for subkey in original[key]:
                    self.assertTrue(subkey in generated[key])


if __name__ == '__main__':
    unittest.main()





class BOOL:
    """  Boolean container -- supports global boolean variables """
    def __init__(self, v: bool) -> None:
        self.v = v

    def __bool__(self):
        return self.v

    def __str__(self):
        return str(self.v)


GLOBAL_STRICT = BOOL(True)


def strict() -> bool:
    """ Switch to global strict mode """
    rval = GLOBAL_STRICT.v
    GLOBAL_STRICT.v = True
    return rval


def lax() -> bool:
    """ Switch to global lax mode """
    rval = GLOBAL_STRICT.v
    GLOBAL_STRICT.v = False
    return rval


def is_strict() -> bool:
    """ Return the global strictness setting """
    return bool(GLOBAL_STRICT)


# Auto generated from .yaml by pythongen.py version:
# Generation date:
# Schema: ifabsent
#
# id: http://example.org/tests/ifabsent
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
from biolinkml.utils.metamodelcore import Bool, ElementIdentifier, NCName, NodeIdentifier, URI, URIorCURIE, XSDDate, XSDDateTime, XSDTime

metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
SHEX = CurieNamespace('shex', 'http://www.w3.org/ns/shex#')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
TEST = CurieNamespace('test', 'http://example.org/test/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = TEST


# Types
class String(str):
    """ A character string """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = TEST.String


class Integer(int):
    """ An integer """
    type_class_uri = XSD.integer
    type_class_curie = "xsd:integer"
    type_name = "integer"
    type_model_uri = TEST.Integer


class Boolean(Bool):
    """ A binary (true or false) value """
    type_class_uri = XSD.boolean
    type_class_curie = "xsd:boolean"
    type_name = "boolean"
    type_model_uri = TEST.Boolean


class Float(float):
    """ A real number that conforms to the xsd:float specification """
    type_class_uri = XSD.float
    type_class_curie = "xsd:float"
    type_name = "float"
    type_model_uri = TEST.Float


class Double(float):
    """ A real number that conforms to the xsd:double specification """
    type_class_uri = XSD.double
    type_class_curie = "xsd:double"
    type_name = "double"
    type_model_uri = TEST.Double


class Time(XSDTime):
    """ A time object represents a (local) time of day, independent of any particular day """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "time"
    type_model_uri = TEST.Time


class Date(XSDDate):
    """ a date (year, month and day) in an idealized calendar """
    type_class_uri = XSD.date
    type_class_curie = "xsd:date"
    type_name = "date"
    type_model_uri = TEST.Date


class Datetime(XSDDateTime):
    """ The combination of a date and time """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "datetime"
    type_model_uri = TEST.Datetime


class Uriorcurie(URIorCURIE):
    """ a URI or a CURIE """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uriorcurie"
    type_model_uri = TEST.Uriorcurie


class Uri(URI):
    """ a complete URI """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uri"
    type_model_uri = TEST.Uri


class Ncname(NCName):
    """ Prefix part of CURIE """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "ncname"
    type_model_uri = TEST.Ncname


class Objectidentifier(ElementIdentifier):
    """ A URI or CURIE that represents an object in the model. """
    type_class_uri = SHEX.iri
    type_class_curie = "shex:iri"
    type_name = "objectidentifier"
    type_model_uri = TEST.Objectidentifier


class Nodeidentifier(NodeIdentifier):
    """ A URI, CURIE or BNODE that represents a node in a model. """
    type_class_uri = SHEX.nonLiteral
    type_class_curie = "shex:nonLiteral"
    type_name = "nodeidentifier"
    type_model_uri = TEST.Nodeidentifier


# Class references



@dataclass
class C1(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = TEST.C1
    class_class_curie: ClassVar[str] = "test:C1"
    class_name: ClassVar[str] = "c1"
    class_model_uri: ClassVar[URIRef] = TEST.C1

    s1: Optional[str] = True
    s1p: Optional[str] = True
    s2: Optional[str] = False
    s2p: Optional[str] = False
    slot_uri: Optional[str] = None
    slot_curie: Optional[str] = None
    class_uri: Optional[str] = None
    class_curie: Optional[str] = None
    bnode: Optional[str] = bnode()
    txt: Optional[str] = "penguins\"doves"
    int: Optional[str] = -1403
    dfltrange: Optional[str] = None
    dfltns: Optional[str] = None


# Slots
class slots:
    pass

slots.s1 = Slot(uri=TEST.s1, name="s1", curie=TEST.curie('s1'),
                      model_uri=TEST.s1, domain=None, range=Optional[str])

slots.s1p = Slot(uri=TEST.s1p, name="s1p", curie=TEST.curie('s1p'),
                      model_uri=TEST.s1p, domain=None, range=Optional[str])

slots.s2 = Slot(uri=TEST.s2, name="s2", curie=TEST.curie('s2'),
                      model_uri=TEST.s2, domain=None, range=Optional[str])

slots.s2p = Slot(uri=TEST.s2p, name="s2p", curie=TEST.curie('s2p'),
                      model_uri=TEST.s2p, domain=None, range=Optional[str])

slots.slot_uri = Slot(uri=TEST.slot_uri, name="slot_uri", curie=TEST.curie('slot_uri'),
                      model_uri=TEST.slot_uri, domain=None, range=Optional[str])

slots.slot_curie = Slot(uri=TEST.slot_curie, name="slot_curie", curie=TEST.curie('slot_curie'),
                      model_uri=TEST.slot_curie, domain=None, range=Optional[str])

slots.class_uri = Slot(uri=TEST.class_uri, name="class_uri", curie=TEST.curie('class_uri'),
                      model_uri=TEST.class_uri, domain=None, range=Optional[str])

slots.class_curie = Slot(uri=TEST.class_curie, name="class_curie", curie=TEST.curie('class_curie'),
                      model_uri=TEST.class_curie, domain=None, range=Optional[str])

slots.bnode = Slot(uri=TEST.bnode, name="bnode", curie=TEST.curie('bnode'),
                      model_uri=TEST.bnode, domain=None, range=Optional[str])

slots.txt = Slot(uri=TEST.txt, name="txt", curie=TEST.curie('txt'),
                      model_uri=TEST.txt, domain=None, range=Optional[str])

slots.int = Slot(uri=TEST.int, name="int", curie=TEST.curie('int'),
                      model_uri=TEST.int, domain=None, range=Optional[str])

slots.dfltrange = Slot(uri=TEST.dfltrange, name="dfltrange", curie=TEST.curie('dfltrange'),
                      model_uri=TEST.dfltrange, domain=None, range=Optional[str])

slots.dfltns = Slot(uri=TEST.dfltns, name="dfltns", curie=TEST.curie('dfltns'),
                      model_uri=TEST.dfltns, domain=None, range=Optional[str])


import copy
import os
import time
from datetime import datetime
from io import StringIO
from typing import Union, TextIO, Optional
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import Request, urlopen

import yaml

from biolinkml.meta import SchemaDefinition, metamodel_version
from biolinkml.utils.mergeutils import merge_schemas, set_from_schema
from biolinkml.utils.namespaces import Namespaces
from biolinkml.utils.yamlutils import DupCheckYamlLoader


def load_raw_schema(data: Union[str, dict, TextIO],
                    source_file: Optional[str] = None,
                    source_file_date: Optional[str] = None,
                    source_file_size: Optional[int] = None,
                    base_dir: Optional[str] = None) -> SchemaDefinition:
    """ Load and flatten SchemaDefinition from a file name, a URL or a block of text

    @param data: URL, file name or block of text YAML Object or open file handle
    @param source_file: Source file name for the schema if data is type TextIO
    @param source_file_date: timestamp of source file if data is type TextIO
    @param source_file_size: size of source file if data is type TextIO
    @param base_dir: Working directory or base URL of sources
    @return: Un-processed Schema Definition object
    """
    def _name_from_url(url) -> str:
        return urlparse(url).path.rsplit('/', 1)[-1].rsplit('.', 1)[0]

    if isinstance(data, str):
        # If passing the actual YAML
        if '\n' in data:
            return load_raw_schema(StringIO(data), source_file, source_file_date, source_file_size, base_dir)

        # Passing a URL or file name
        assert source_file is None, "source_file parameter not allowed if data is a file or URL"
        assert source_file_date is None, "source_file_date parameter not allowed if data is a file or URL"
        assert source_file_size is None, "source_file_size parameter not allowed if data is a file or URL"

        if '://' in data or (base_dir and '://' in base_dir):
            # URL
            fname = Namespaces.join(base_dir, data) if '://' not in data else data
            req = Request(fname)
            req.add_header("Accept", "text/yaml, application/yaml;q=0.9")
            try:
                response = urlopen(req)
            except HTTPError as e:
                # This is here because the message out of urllib doesn't include the file name
                e.msg = f"{e.filename}"
                raise e
            with response:
                return load_raw_schema(response, fname, response.info()['Last-Modified'],
                                       response.info()['Content-Length'])


        else:
            # File name
            if not base_dir:
                fname = os.path.abspath(data)
                base_dir = os.path.dirname(fname)
            else:
                fname = data if os.path.isabs(data) else os.path.abspath(os.path.join(base_dir, data))
            with open(fname) as f:
                return load_raw_schema(f, fname, time.ctime(os.path.getmtime(fname)), os.path.getsize(fname), base_dir)
    else:
        # Loaded YAML or file handle that references YAML
        schemadefs = copy.deepcopy(data) if isinstance(data, dict) else yaml.load(data, DupCheckYamlLoader)

        # Convert the schema into a "name: definition" form
        if not all(isinstance(e, dict) for e in schemadefs.values()):
            if 'name' in schemadefs:
                schemaname = schemadefs.pop('name')
            elif 'id' in schemadefs:
                schemaname = _name_from_url(schemadefs['id'])
            else:
                raise ValueError("Unable to determine schema name")
            schema_body = [schemadefs]
            schemadefs = {schemaname: schemadefs}
        else:
            schema_body = list(schemadefs.values())

        def check_is_dict(element: str) -> None:
            """ Verify that element is an instance of a dictionary, mapping empty elements to dictionaries """
            for schemaname, body in schemadefs.items():
                if element in body:
                    if body[element] is None:
                        body[element] = dict()
                    elif not isinstance(body[element], dict):
                        raise ValueError(f'Schema: {schemaname} - Element: {element} must be a dictionary')

        def fix_multiples(container:  str, element: str) -> None:
            """
            A common error is representing a list object as a singleton.  This fixes this problem
            :param container: name of container to fix (e.g. a specific clas instance)
            :param element:  name or list element to adjust (e.g. notes"
            """
            # Note: multiple bodies in the schema are an at-risk feature.  Doesn't seem to have a real use case.
            for body in schema_body:
                if container in body:
                    for c in body[container].values():
                        if c and element in c and isinstance(c[element], str):
                            c[element] = [c[element]]

        for e in ['slots', 'classes', 'types', 'subsets']:
            """ Validate the basic categories, fixing multiples where appropriate """
            check_is_dict(e)
            fix_multiples(e, 'in_subset')
            fix_multiples(e, 'apply_to')

        for e in ['imports']:
            for body in schema_body:
                if e in body:
                    if isinstance(body[e], str):
                        body[e] = [body[e]]

        # Add the implicit domain to the slot usages
        for body in schema_body:
            for cname, cls in body.get('classes', {}).items():
                if cls is None:
                    cls = {}
                    body['classes'][cname] = cls
                for uname, usage in cls.get('slot usage', {}).items():
                    if usage is None:
                        usage = {}
                        cls['slot usage'][uname] = usage
                    if 'domain' not in usage:
                        usage['domain'] = cname

        schema: SchemaDefinition = None
        for sname, sdef in {k: SchemaDefinition(name=k, **v) for k, v in schemadefs.items()}.items():
            if schema is None:
                schema = sdef
                if source_file:
                    schema.source_file = source_file
                schema.source_file_date = source_file_date
                schema.source_file_size = source_file_size
                schema.generation_date = datetime.now().strftime("%Y-%m-%d %H:%M")
                schema.metamodel_version = metamodel_version
                set_from_schema(schema)
            else:
                merge_schemas(schema, sdef)
        return schema



# Auto generated from meta.yaml by pythongen.py version: 0.4.0
# Generation date: 2020-06-07 17:46
# Schema: metamodel
#
# id: https://w3id.org/biolink/biolinkml/meta
# description: A metamodel for defining biolink related schemas
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
from biolinkml.utils.metamodelcore import Bool, NCName, URI, URIorCURIE, XSDDateTime
from includes.types import Boolean, Datetime, Integer, Ncname, String, Uri, Uriorcurie

metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
OIO = CurieNamespace('OIO', 'http://www.geneontology.org/formats/oboInOwl#')
BIOLINKML = CurieNamespace('biolinkml', 'https://w3id.org/biolink/biolinkml/')
DCTERMS = CurieNamespace('dcterms', 'http://purl.org/dc/terms/')
META = CurieNamespace('meta', 'https://w3id.org/biolink/biolinkml/meta/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = META


# Types

# Class references
class ElementName(extended_str):
    pass


class SchemaDefinitionName(ElementName):
    pass


class TypeDefinitionName(ElementName):
    pass


class SubsetDefinitionName(ElementName):
    pass


class DefinitionName(ElementName):
    pass


class SlotDefinitionName(DefinitionName):
    pass


class ClassDefinitionName(DefinitionName):
    pass


class PrefixPrefixPrefix(NCName):
    pass


class LocalNameLocalNameSource(NCName):
    pass


class AltDescriptionSource(NCName):
    pass


@dataclass
class Element(YAMLRoot):
    """
    a named element in the model
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.Element
    class_class_curie: ClassVar[str] = "meta:Element"
    class_name: ClassVar[str] = "element"
    class_model_uri: ClassVar[URIRef] = META.Element

    name: Union[str, ElementName]
    id_prefixes: List[Union[str, NCName]] = empty_list()
    definition_uri: Optional[Union[str, URIorCURIE]] = None
    aliases: List[str] = empty_list()
    local_names: Union[dict, "LocalName"] = empty_dict()
    mappings: List[Union[str, URIorCURIE]] = empty_list()
    description: Optional[str] = None
    alt_descriptions: Union[dict, "AltDescription"] = empty_dict()
    deprecated: Optional[str] = None
    todos: List[str] = empty_list()
    notes: List[str] = empty_list()
    comments: List[str] = empty_list()
    examples: List[Union[dict, "Example"]] = empty_list()
    in_subset: List[Union[str, SubsetDefinitionName]] = empty_list()
    from_schema: Optional[Union[str, URI]] = None
    imported_from: Optional[str] = None
    see_also: List[Union[str, URIorCURIE]] = empty_list()
    exact_mappings: List[Union[str, URIorCURIE]] = empty_list()
    close_mappings: List[Union[str, URIorCURIE]] = empty_list()
    related_mappings: List[Union[str, URIorCURIE]] = empty_list()
    deprecated_element_has_exact_replacement: Optional[Union[str, URIorCURIE]] = None
    deprecated_element_has_possible_replacement: Optional[Union[str, URIorCURIE]] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        self.id_prefixes = [v if isinstance(v, NCName)
                            else NCName(v) for v in self.id_prefixes]
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, ElementName):
            self.name = ElementName(self.name)
        if self.definition_uri is not None and not isinstance(self.definition_uri, URIorCURIE):
            self.definition_uri = URIorCURIE(self.definition_uri)
        for k, v in self.local_names.items():
            if not isinstance(v, LocalName):
                self.local_names[k] = LocalName(k, v)
        self.mappings = [v if isinstance(v, URIorCURIE)
                         else URIorCURIE(v) for v in self.mappings]
        for k, v in self.alt_descriptions.items():
            if not isinstance(v, AltDescription):
                self.alt_descriptions[k] = AltDescription(k, v)
        self.examples = [v if isinstance(v, Example)
                         else Example(**v) for v in self.examples]
        self.in_subset = [v if isinstance(v, SubsetDefinitionName)
                          else SubsetDefinitionName(v) for v in self.in_subset]
        if self.from_schema is not None and not isinstance(self.from_schema, URI):
            self.from_schema = URI(self.from_schema)
        self.see_also = [v if isinstance(v, URIorCURIE)
                         else URIorCURIE(v) for v in self.see_also]
        self.exact_mappings = [v if isinstance(v, URIorCURIE)
                               else URIorCURIE(v) for v in self.exact_mappings]
        self.close_mappings = [v if isinstance(v, URIorCURIE)
                               else URIorCURIE(v) for v in self.close_mappings]
        self.related_mappings = [v if isinstance(v, URIorCURIE)
                                 else URIorCURIE(v) for v in self.related_mappings]
        if self.deprecated_element_has_exact_replacement is not None and not isinstance(self.deprecated_element_has_exact_replacement, URIorCURIE):
            self.deprecated_element_has_exact_replacement = URIorCURIE(self.deprecated_element_has_exact_replacement)
        if self.deprecated_element_has_possible_replacement is not None and not isinstance(self.deprecated_element_has_possible_replacement, URIorCURIE):
            self.deprecated_element_has_possible_replacement = URIorCURIE(self.deprecated_element_has_possible_replacement)
        super().__post_init__(**kwargs)


@dataclass
class SchemaDefinition(Element):
    """
    a collection of subset, type, slot and class definitions
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.SchemaDefinition
    class_class_curie: ClassVar[str] = "meta:SchemaDefinition"
    class_name: ClassVar[str] = "schema_definition"
    class_model_uri: ClassVar[URIRef] = META.SchemaDefinition

    name: Union[str, SchemaDefinitionName] = None
    id: Union[str, URI] = None
    title: Optional[str] = None
    version: Optional[str] = None
    imports: List[Union[str, URIorCURIE]] = empty_list()
    license: Optional[str] = None
    prefixes: Union[dict, "Prefix"] = empty_dict()
    emit_prefixes: List[Union[str, NCName]] = empty_list()
    default_curi_maps: List[str] = empty_list()
    default_prefix: Optional[str] = None
    default_range: Optional[Union[str, TypeDefinitionName]] = None
    subsets: Dict[Union[str, SubsetDefinitionName], Union[dict, "SubsetDefinition"]] = empty_dict()
    types: Dict[Union[str, TypeDefinitionName], Union[dict, "TypeDefinition"]] = empty_dict()
    slots: Dict[Union[str, SlotDefinitionName], Union[dict, "SlotDefinition"]] = empty_dict()
    classes: Dict[Union[str, ClassDefinitionName], Union[dict, "ClassDefinition"]] = empty_dict()
    metamodel_version: Optional[str] = None
    source_file: Optional[str] = None
    source_file_date: Optional[Union[str, XSDDateTime]] = None
    source_file_size: Optional[int] = None
    generation_date: Optional[Union[str, XSDDateTime]] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.default_prefix is None:
            self.default_prefix = sfx(str(self.id))
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, SchemaDefinitionName):
            self.name = SchemaDefinitionName(self.name)
        if self.id is None:
            raise ValueError(f"id must be supplied")
        if not isinstance(self.id, URI):
            self.id = URI(self.id)
        self.imports = [v if isinstance(v, URIorCURIE)
                        else URIorCURIE(v) for v in self.imports]
        for k, v in self.prefixes.items():
            if not isinstance(v, Prefix):
                self.prefixes[k] = Prefix(k, v)
        self.emit_prefixes = [v if isinstance(v, NCName)
                              else NCName(v) for v in self.emit_prefixes]
        if self.default_range is not None and not isinstance(self.default_range, TypeDefinitionName):
            self.default_range = TypeDefinitionName(self.default_range)
        for k, v in self.subsets.items():
            if not isinstance(v, SubsetDefinition):
                self.subsets[k] = SubsetDefinition(name=k, **({} if v is None else v))
        for k, v in self.types.items():
            if not isinstance(v, TypeDefinition):
                self.types[k] = TypeDefinition(name=k, **({} if v is None else v))
        for k, v in self.slots.items():
            if not isinstance(v, SlotDefinition):
                self.slots[k] = SlotDefinition(name=k, **({} if v is None else v))
        for k, v in self.classes.items():
            if not isinstance(v, ClassDefinition):
                self.classes[k] = ClassDefinition(name=k, **({} if v is None else v))
        if self.source_file_date is not None and not isinstance(self.source_file_date, XSDDateTime):
            self.source_file_date = XSDDateTime(self.source_file_date)
        if self.generation_date is not None and not isinstance(self.generation_date, XSDDateTime):
            self.generation_date = XSDDateTime(self.generation_date)
        super().__post_init__(**kwargs)


@dataclass
class TypeDefinition(Element):
    """
    A data type definition.
    """
    _inherited_slots: ClassVar[List[str]] = ["base", "uri", "repr"]

    class_class_uri: ClassVar[URIRef] = META.TypeDefinition
    class_class_curie: ClassVar[str] = "meta:TypeDefinition"
    class_name: ClassVar[str] = "type_definition"
    class_model_uri: ClassVar[URIRef] = META.TypeDefinition

    name: Union[str, TypeDefinitionName] = None
    typeof: Optional[Union[str, TypeDefinitionName]] = None
    base: Optional[str] = None
    uri: Optional[Union[str, URIorCURIE]] = None
    repr: Optional[str] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, TypeDefinitionName):
            self.name = TypeDefinitionName(self.name)
        if self.typeof is not None and not isinstance(self.typeof, TypeDefinitionName):
            self.typeof = TypeDefinitionName(self.typeof)
        if self.uri is not None and not isinstance(self.uri, URIorCURIE):
            self.uri = URIorCURIE(self.uri)
        super().__post_init__(**kwargs)


@dataclass
class SubsetDefinition(Element):
    """
    the name and description of a subset
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.SubsetDefinition
    class_class_curie: ClassVar[str] = "meta:SubsetDefinition"
    class_name: ClassVar[str] = "subset_definition"
    class_model_uri: ClassVar[URIRef] = META.SubsetDefinition

    name: Union[str, SubsetDefinitionName] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, SubsetDefinitionName):
            self.name = SubsetDefinitionName(self.name)
        super().__post_init__(**kwargs)


@dataclass
class Definition(Element):
    """
    base class for definitions
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.Definition
    class_class_curie: ClassVar[str] = "meta:Definition"
    class_name: ClassVar[str] = "definition"
    class_model_uri: ClassVar[URIRef] = META.Definition

    name: Union[str, DefinitionName] = None
    is_a: Optional[Union[str, DefinitionName]] = None
    abstract: Optional[Bool] = None
    mixin: Optional[Bool] = None
    mixins: List[Union[str, DefinitionName]] = empty_list()
    apply_to: List[Union[str, DefinitionName]] = empty_list()
    values_from: List[Union[str, URIorCURIE]] = empty_list()

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.is_a is not None and not isinstance(self.is_a, DefinitionName):
            self.is_a = DefinitionName(self.is_a)
        self.mixins = [v if isinstance(v, DefinitionName)
                       else DefinitionName(v) for v in self.mixins]
        self.apply_to = [v if isinstance(v, DefinitionName)
                         else DefinitionName(v) for v in self.apply_to]
        self.values_from = [v if isinstance(v, URIorCURIE)
                            else URIorCURIE(v) for v in self.values_from]
        super().__post_init__(**kwargs)


@dataclass
class SlotDefinition(Definition):
    """
    the definition of a property or a slot
    """
    _inherited_slots: ClassVar[List[str]] = ["domain", "range", "multivalued", "inherited", "readonly", "ifabsent", "required", "inlined", "key", "identifier", "role"]

    class_class_uri: ClassVar[URIRef] = META.SlotDefinition
    class_class_curie: ClassVar[str] = "meta:SlotDefinition"
    class_name: ClassVar[str] = "slot_definition"
    class_model_uri: ClassVar[URIRef] = META.SlotDefinition

    name: Union[str, SlotDefinitionName] = None
    is_a: Optional[Union[str, SlotDefinitionName]] = None
    mixins: List[Union[str, SlotDefinitionName]] = empty_list()
    apply_to: List[Union[str, SlotDefinitionName]] = empty_list()
    singular_name: Optional[str] = None
    domain: Optional[Union[str, ClassDefinitionName]] = None
    range: Optional[Union[str, ElementName]] = None
    slot_uri: Optional[Union[str, URIorCURIE]] = None
    multivalued: Optional[Bool] = None
    inherited: Optional[Bool] = None
    readonly: Optional[str] = None
    ifabsent: Optional[str] = None
    required: Optional[Bool] = None
    inlined: Optional[Bool] = None
    key: Optional[Bool] = None
    identifier: Optional[Bool] = None
    alias: Optional[str] = None
    owner: Optional[Union[str, DefinitionName]] = None
    subproperty_of: Optional[Union[str, URIorCURIE]] = None
    symmetric: Optional[Bool] = None
    inverse: Optional[Union[str, SlotDefinitionName]] = None
    is_class_field: Optional[Bool] = None
    role: Optional[str] = None
    is_usage_slot: Optional[Bool] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, SlotDefinitionName):
            self.name = SlotDefinitionName(self.name)
        if self.is_a is not None and not isinstance(self.is_a, SlotDefinitionName):
            self.is_a = SlotDefinitionName(self.is_a)
        self.mixins = [v if isinstance(v, SlotDefinitionName)
                       else SlotDefinitionName(v) for v in self.mixins]
        self.apply_to = [v if isinstance(v, SlotDefinitionName)
                         else SlotDefinitionName(v) for v in self.apply_to]
        if self.domain is not None and not isinstance(self.domain, ClassDefinitionName):
            self.domain = ClassDefinitionName(self.domain)
        if self.range is not None and not isinstance(self.range, ElementName):
            self.range = ElementName(self.range)
        if self.slot_uri is not None and not isinstance(self.slot_uri, URIorCURIE):
            self.slot_uri = URIorCURIE(self.slot_uri)
        if self.owner is not None and not isinstance(self.owner, DefinitionName):
            self.owner = DefinitionName(self.owner)
        if self.subproperty_of is not None and not isinstance(self.subproperty_of, URIorCURIE):
            self.subproperty_of = URIorCURIE(self.subproperty_of)
        if self.inverse is not None and not isinstance(self.inverse, SlotDefinitionName):
            self.inverse = SlotDefinitionName(self.inverse)
        super().__post_init__(**kwargs)


@dataclass
class ClassDefinition(Definition):
    """
    the definition of a class or interface
    """
    _inherited_slots: ClassVar[List[str]] = ["defining_slots"]

    class_class_uri: ClassVar[URIRef] = META.ClassDefinition
    class_class_curie: ClassVar[str] = "meta:ClassDefinition"
    class_name: ClassVar[str] = "class_definition"
    class_model_uri: ClassVar[URIRef] = META.ClassDefinition

    name: Union[str, ClassDefinitionName] = None
    is_a: Optional[Union[str, ClassDefinitionName]] = None
    mixins: List[Union[str, ClassDefinitionName]] = empty_list()
    apply_to: List[Union[str, ClassDefinitionName]] = empty_list()
    slots: List[Union[str, SlotDefinitionName]] = empty_list()
    slot_usage: Dict[Union[str, SlotDefinitionName], Union[dict, SlotDefinition]] = empty_dict()
    class_uri: Optional[Union[str, URIorCURIE]] = None
    subclass_of: Optional[Union[str, URIorCURIE]] = None
    union_of: List[Union[str, ClassDefinitionName]] = empty_list()
    defining_slots: List[Union[str, SlotDefinitionName]] = empty_list()

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.name is None:
            raise ValueError(f"name must be supplied")
        if not isinstance(self.name, ClassDefinitionName):
            self.name = ClassDefinitionName(self.name)
        if self.is_a is not None and not isinstance(self.is_a, ClassDefinitionName):
            self.is_a = ClassDefinitionName(self.is_a)
        self.mixins = [v if isinstance(v, ClassDefinitionName)
                       else ClassDefinitionName(v) for v in self.mixins]
        self.apply_to = [v if isinstance(v, ClassDefinitionName)
                         else ClassDefinitionName(v) for v in self.apply_to]
        self.slots = [v if isinstance(v, SlotDefinitionName)
                      else SlotDefinitionName(v) for v in self.slots]
        for k, v in self.slot_usage.items():
            if not isinstance(v, SlotDefinition):
                self.slot_usage[k] = SlotDefinition(name=k, **({} if v is None else v))
        if self.class_uri is not None and not isinstance(self.class_uri, URIorCURIE):
            self.class_uri = URIorCURIE(self.class_uri)
        if self.subclass_of is not None and not isinstance(self.subclass_of, URIorCURIE):
            self.subclass_of = URIorCURIE(self.subclass_of)
        self.union_of = [v if isinstance(v, ClassDefinitionName)
                         else ClassDefinitionName(v) for v in self.union_of]
        self.defining_slots = [v if isinstance(v, SlotDefinitionName)
                               else SlotDefinitionName(v) for v in self.defining_slots]
        super().__post_init__(**kwargs)


@dataclass
class Prefix(YAMLRoot):
    """
    prefix URI tuple
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.Prefix
    class_class_curie: ClassVar[str] = "meta:Prefix"
    class_name: ClassVar[str] = "prefix"
    class_model_uri: ClassVar[URIRef] = META.Prefix

    prefix_prefix: Union[str, PrefixPrefixPrefix]
    prefix_reference: Union[str, URI]

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.prefix_prefix is None:
            raise ValueError(f"prefix_prefix must be supplied")
        if not isinstance(self.prefix_prefix, PrefixPrefixPrefix):
            self.prefix_prefix = PrefixPrefixPrefix(self.prefix_prefix)
        if self.prefix_reference is None:
            raise ValueError(f"prefix_reference must be supplied")
        if not isinstance(self.prefix_reference, URI):
            self.prefix_reference = URI(self.prefix_reference)
        super().__post_init__(**kwargs)


@dataclass
class LocalName(YAMLRoot):
    """
    an attributed label
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.LocalName
    class_class_curie: ClassVar[str] = "meta:LocalName"
    class_name: ClassVar[str] = "local_name"
    class_model_uri: ClassVar[URIRef] = META.LocalName

    local_name_source: Union[str, LocalNameLocalNameSource]
    local_name_value: str

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.local_name_source is None:
            raise ValueError(f"local_name_source must be supplied")
        if not isinstance(self.local_name_source, LocalNameLocalNameSource):
            self.local_name_source = LocalNameLocalNameSource(self.local_name_source)
        if self.local_name_value is None:
            raise ValueError(f"local_name_value must be supplied")
        super().__post_init__(**kwargs)


@dataclass
class Example(YAMLRoot):
    """
    usage example and description
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.Example
    class_class_curie: ClassVar[str] = "meta:Example"
    class_name: ClassVar[str] = "example"
    class_model_uri: ClassVar[URIRef] = META.Example

    value: Optional[str] = None
    description: Optional[str] = None

@dataclass
class AltDescription(YAMLRoot):
    """
    an attributed description
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = META.AltDescription
    class_class_curie: ClassVar[str] = "meta:AltDescription"
    class_name: ClassVar[str] = "alt_description"
    class_model_uri: ClassVar[URIRef] = META.AltDescription

    source: Union[str, AltDescriptionSource]
    description: str

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.source is None:
            raise ValueError(f"source must be supplied")
        if not isinstance(self.source, AltDescriptionSource):
            self.source = AltDescriptionSource(self.source)
        if self.description is None:
            raise ValueError(f"description must be supplied")
        super().__post_init__(**kwargs)



# Slots
class slots:
    pass

slots.name = Slot(uri=RDFS.label, name="name", curie=RDFS.curie('label'),
                      model_uri=META.name, domain=Element, range=Union[str, ElementName])

slots.definition_uri = Slot(uri=META.definition_uri, name="definition_uri", curie=META.curie('definition_uri'),
                      model_uri=META.definition_uri, domain=Element, range=Optional[Union[str, URIorCURIE]])

slots.id_prefixes = Slot(uri=META.id_prefixes, name="id_prefixes", curie=META.curie('id_prefixes'),
                      model_uri=META.id_prefixes, domain=Element, range=List[Union[str, NCName]])

slots.description = Slot(uri=SKOS.definition, name="description", curie=SKOS.curie('definition'),
                      model_uri=META.description, domain=Element, range=Optional[str])

slots.aliases = Slot(uri=SKOS.altLabel, name="aliases", curie=SKOS.curie('altLabel'),
                      model_uri=META.aliases, domain=Element, range=List[str])

slots.deprecated = Slot(uri=META.deprecated, name="deprecated", curie=META.curie('deprecated'),
                      model_uri=META.deprecated, domain=Element, range=Optional[str])

slots.todos = Slot(uri=META.todos, name="todos", curie=META.curie('todos'),
                      model_uri=META.todos, domain=Element, range=List[str])

slots.notes = Slot(uri=SKOS.editorialNote, name="notes", curie=SKOS.curie('editorialNote'),
                      model_uri=META.notes, domain=Element, range=List[str])

slots.comments = Slot(uri=SKOS.note, name="comments", curie=SKOS.curie('note'),
                      model_uri=META.comments, domain=Element, range=List[str])

slots.in_subset = Slot(uri=OIO.inSubset, name="in_subset", curie=OIO.curie('inSubset'),
                      model_uri=META.in_subset, domain=Element, range=List[Union[str, SubsetDefinitionName]])

slots.from_schema = Slot(uri=SKOS.inScheme, name="from_schema", curie=SKOS.curie('inScheme'),
                      model_uri=META.from_schema, domain=Element, range=Optional[Union[str, URI]])

slots.imported_from = Slot(uri=META.imported_from, name="imported_from", curie=META.curie('imported_from'),
                      model_uri=META.imported_from, domain=Element, range=Optional[str])

slots.see_also = Slot(uri=RDFS.seeAlso, name="see_also", curie=RDFS.curie('seeAlso'),
                      model_uri=META.see_also, domain=Element, range=List[Union[str, URIorCURIE]])

slots.is_a = Slot(uri=META.is_a, name="is_a", curie=META.curie('is_a'),
                      model_uri=META.is_a, domain=Definition, range=Optional[Union[str, DefinitionName]])

slots.abstract = Slot(uri=META.abstract, name="abstract", curie=META.curie('abstract'),
                      model_uri=META.abstract, domain=Definition, range=Optional[Bool])

slots.mixin = Slot(uri=META.mixin, name="mixin", curie=META.curie('mixin'),
                      model_uri=META.mixin, domain=Definition, range=Optional[Bool])

slots.mixins = Slot(uri=META.mixins, name="mixins", curie=META.curie('mixins'),
                      model_uri=META.mixins, domain=Definition, range=List[Union[str, DefinitionName]])

slots.apply_to = Slot(uri=META.apply_to, name="apply_to", curie=META.curie('apply_to'),
                      model_uri=META.apply_to, domain=Definition, range=List[Union[str, DefinitionName]])

slots.values_from = Slot(uri=META.values_from, name="values_from", curie=META.curie('values_from'),
                      model_uri=META.values_from, domain=Definition, range=List[Union[str, URIorCURIE]])

slots.id = Slot(uri=META.id, name="id", curie=META.curie('id'),
                      model_uri=META.id, domain=SchemaDefinition, range=Union[str, URI])

slots.emit_prefixes = Slot(uri=META.emit_prefixes, name="emit_prefixes", curie=META.curie('emit_prefixes'),
                      model_uri=META.emit_prefixes, domain=SchemaDefinition, range=List[Union[str, NCName]])

slots.title = Slot(uri=DCTERMS.title, name="title", curie=DCTERMS.curie('title'),
                      model_uri=META.title, domain=SchemaDefinition, range=Optional[str])

slots.version = Slot(uri=PAV.version, name="version", curie=PAV.curie('version'),
                      model_uri=META.version, domain=SchemaDefinition, range=Optional[str])

slots.imports = Slot(uri=META.imports, name="imports", curie=META.curie('imports'),
                      model_uri=META.imports, domain=SchemaDefinition, range=List[Union[str, URIorCURIE]])

slots.license = Slot(uri=DCTERMS.license, name="license", curie=DCTERMS.curie('license'),
                      model_uri=META.license, domain=SchemaDefinition, range=Optional[str])

slots.default_curi_maps = Slot(uri=META.default_curi_maps, name="default_curi_maps", curie=META.curie('default_curi_maps'),
                      model_uri=META.default_curi_maps, domain=SchemaDefinition, range=List[str])

slots.default_prefix = Slot(uri=META.default_prefix, name="default_prefix", curie=META.curie('default_prefix'),
                      model_uri=META.default_prefix, domain=SchemaDefinition, range=Optional[str])

slots.default_range = Slot(uri=META.default_range, name="default_range", curie=META.curie('default_range'),
                      model_uri=META.default_range, domain=SchemaDefinition, range=Optional[Union[str, TypeDefinitionName]])

slots.subsets = Slot(uri=META.subsets, name="subsets", curie=META.curie('subsets'),
                      model_uri=META.subsets, domain=SchemaDefinition, range=Dict[Union[str, SubsetDefinitionName], Union[dict, "SubsetDefinition"]])

slots.types = Slot(uri=META.types, name="types", curie=META.curie('types'),
                      model_uri=META.types, domain=SchemaDefinition, range=Dict[Union[str, TypeDefinitionName], Union[dict, "TypeDefinition"]])

slots.slot_definitions = Slot(uri=META.slots, name="slot_definitions", curie=META.curie('slots'),
                      model_uri=META.slot_definitions, domain=SchemaDefinition, range=Dict[Union[str, SlotDefinitionName], Union[dict, "SlotDefinition"]])

slots.classes = Slot(uri=META.classes, name="classes", curie=META.curie('classes'),
                      model_uri=META.classes, domain=SchemaDefinition, range=Dict[Union[str, ClassDefinitionName], Union[dict, "ClassDefinition"]])

slots.metamodel_version = Slot(uri=META.metamodel_version, name="metamodel_version", curie=META.curie('metamodel_version'),
                      model_uri=META.metamodel_version, domain=SchemaDefinition, range=Optional[str])

slots.source_file = Slot(uri=META.source_file, name="source_file", curie=META.curie('source_file'),
                      model_uri=META.source_file, domain=SchemaDefinition, range=Optional[str])

slots.source_file_date = Slot(uri=META.source_file_date, name="source_file_date", curie=META.curie('source_file_date'),
                      model_uri=META.source_file_date, domain=SchemaDefinition, range=Optional[Union[str, XSDDateTime]])

slots.source_file_size = Slot(uri=META.source_file_size, name="source_file_size", curie=META.curie('source_file_size'),
                      model_uri=META.source_file_size, domain=SchemaDefinition, range=Optional[int])

slots.generation_date = Slot(uri=META.generation_date, name="generation_date", curie=META.curie('generation_date'),
                      model_uri=META.generation_date, domain=SchemaDefinition, range=Optional[Union[str, XSDDateTime]])

slots.slots = Slot(uri=META.slots, name="slots", curie=META.curie('slots'),
                      model_uri=META.slots, domain=ClassDefinition, range=List[Union[str, SlotDefinitionName]])

slots.slot_usage = Slot(uri=META.slot_usage, name="slot_usage", curie=META.curie('slot_usage'),
                      model_uri=META.slot_usage, domain=ClassDefinition, range=Dict[Union[str, SlotDefinitionName], Union[dict, SlotDefinition]])

slots.class_uri = Slot(uri=META.class_uri, name="class_uri", curie=META.curie('class_uri'),
                      model_uri=META.class_uri, domain=ClassDefinition, range=Optional[Union[str, URIorCURIE]])

slots.subclass_of = Slot(uri=RDFS.subClassOf, name="subclass_of", curie=RDFS.curie('subClassOf'),
                      model_uri=META.subclass_of, domain=ClassDefinition, range=Optional[Union[str, URIorCURIE]])

slots.defining_slots = Slot(uri=META.defining_slots, name="defining_slots", curie=META.curie('defining_slots'),
                      model_uri=META.defining_slots, domain=ClassDefinition, range=List[Union[str, SlotDefinitionName]])

slots.union_of = Slot(uri=META.union_of, name="union_of", curie=META.curie('union_of'),
                      model_uri=META.union_of, domain=ClassDefinition, range=List[Union[str, ClassDefinitionName]])

slots.domain = Slot(uri=META.domain, name="domain", curie=META.curie('domain'),
                      model_uri=META.domain, domain=SlotDefinition, range=Optional[Union[str, ClassDefinitionName]])

slots.range = Slot(uri=META.range, name="range", curie=META.curie('range'),
                      model_uri=META.range, domain=SlotDefinition, range=Optional[Union[str, ElementName]])

slots.slot_uri = Slot(uri=META.slot_uri, name="slot_uri", curie=META.curie('slot_uri'),
                      model_uri=META.slot_uri, domain=SlotDefinition, range=Optional[Union[str, URIorCURIE]])

slots.multivalued = Slot(uri=META.multivalued, name="multivalued", curie=META.curie('multivalued'),
                      model_uri=META.multivalued, domain=SlotDefinition, range=Optional[Bool])

slots.inherited = Slot(uri=META.inherited, name="inherited", curie=META.curie('inherited'),
                      model_uri=META.inherited, domain=SlotDefinition, range=Optional[Bool])

slots.readonly = Slot(uri=META.readonly, name="readonly", curie=META.curie('readonly'),
                      model_uri=META.readonly, domain=SlotDefinition, range=Optional[str])

slots.ifabsent = Slot(uri=META.ifabsent, name="ifabsent", curie=META.curie('ifabsent'),
                      model_uri=META.ifabsent, domain=SlotDefinition, range=Optional[str])

slots.singular_name = Slot(uri=SKOS.altLabel, name="singular_name", curie=SKOS.curie('altLabel'),
                      model_uri=META.singular_name, domain=SlotDefinition, range=Optional[str])

slots.required = Slot(uri=META.required, name="required", curie=META.curie('required'),
                      model_uri=META.required, domain=SlotDefinition, range=Optional[Bool])

slots.inlined = Slot(uri=META.inlined, name="inlined", curie=META.curie('inlined'),
                      model_uri=META.inlined, domain=SlotDefinition, range=Optional[Bool])

slots.key = Slot(uri=META.key, name="key", curie=META.curie('key'),
                      model_uri=META.key, domain=SlotDefinition, range=Optional[Bool])

slots.identifier = Slot(uri=META.identifier, name="identifier", curie=META.curie('identifier'),
                      model_uri=META.identifier, domain=SlotDefinition, range=Optional[Bool])

slots.alias = Slot(uri=META.alias, name="alias", curie=META.curie('alias'),
                      model_uri=META.alias, domain=SlotDefinition, range=Optional[str])

slots.owner = Slot(uri=META.owner, name="owner", curie=META.curie('owner'),
                      model_uri=META.owner, domain=SlotDefinition, range=Optional[Union[str, DefinitionName]])

slots.is_usage_slot = Slot(uri=META.is_usage_slot, name="is_usage_slot", curie=META.curie('is_usage_slot'),
                      model_uri=META.is_usage_slot, domain=SlotDefinition, range=Optional[Bool])

slots.subproperty_of = Slot(uri=RDFS.subPropertyOf, name="subproperty_of", curie=RDFS.curie('subPropertyOf'),
                      model_uri=META.subproperty_of, domain=SlotDefinition, range=Optional[Union[str, URIorCURIE]])

slots.symmetric = Slot(uri=META.symmetric, name="symmetric", curie=META.curie('symmetric'),
                      model_uri=META.symmetric, domain=SlotDefinition, range=Optional[Bool])

slots.inverse = Slot(uri=OWL.inverseOf, name="inverse", curie=OWL.curie('inverseOf'),
                      model_uri=META.inverse, domain=SlotDefinition, range=Optional[Union[str, SlotDefinitionName]])

slots.is_class_field = Slot(uri=META.is_class_field, name="is_class_field", curie=META.curie('is_class_field'),
                      model_uri=META.is_class_field, domain=SlotDefinition, range=Optional[Bool])

slots.role = Slot(uri=META.role, name="role", curie=META.curie('role'),
                      model_uri=META.role, domain=SlotDefinition, range=Optional[str])

slots.typeof = Slot(uri=META.typeof, name="typeof", curie=META.curie('typeof'),
                      model_uri=META.typeof, domain=TypeDefinition, range=Optional[Union[str, TypeDefinitionName]])

slots.base = Slot(uri=META.base, name="base", curie=META.curie('base'),
                      model_uri=META.base, domain=TypeDefinition, range=Optional[str])

slots.type_uri = Slot(uri=META.uri, name="type_uri", curie=META.curie('uri'),
                      model_uri=META.type_uri, domain=TypeDefinition, range=Optional[Union[str, URIorCURIE]])

slots.repr = Slot(uri=META.repr, name="repr", curie=META.curie('repr'),
                      model_uri=META.repr, domain=TypeDefinition, range=Optional[str])

slots.alt_description_text = Slot(uri=META.description, name="alt_description_text", curie=META.curie('description'),
                      model_uri=META.alt_description_text, domain=AltDescription, range=str)

slots.alt_description_source = Slot(uri=META.source, name="alt_description_source", curie=META.curie('source'),
                      model_uri=META.alt_description_source, domain=AltDescription, range=Union[str, AltDescriptionSource])

slots.alt_descriptions = Slot(uri=META.alt_descriptions, name="alt_descriptions", curie=META.curie('alt_descriptions'),
                      model_uri=META.alt_descriptions, domain=Element, range=Union[dict, "AltDescription"])

slots.value = Slot(uri=SKOS.example, name="value", curie=SKOS.curie('example'),
                      model_uri=META.value, domain=Example, range=Optional[str])

slots.value_description = Slot(uri=META.description, name="value_description", curie=META.curie('description'),
                      model_uri=META.value_description, domain=Example, range=Optional[str])

slots.examples = Slot(uri=META.examples, name="examples", curie=META.curie('examples'),
                      model_uri=META.examples, domain=Element, range=List[Union[dict, "Example"]])

slots.prefix_prefix = Slot(uri=META.prefix_prefix, name="prefix_prefix", curie=META.curie('prefix_prefix'),
                      model_uri=META.prefix_prefix, domain=Prefix, range=Union[str, PrefixPrefixPrefix])

slots.prefix_reference = Slot(uri=META.prefix_reference, name="prefix_reference", curie=META.curie('prefix_reference'),
                      model_uri=META.prefix_reference, domain=Prefix, range=Union[str, URI])

slots.prefixes = Slot(uri=META.prefixes, name="prefixes", curie=META.curie('prefixes'),
                      model_uri=META.prefixes, domain=SchemaDefinition, range=Union[dict, "Prefix"])

slots.local_name_source = Slot(uri=META.local_name_source, name="local_name_source", curie=META.curie('local_name_source'),
                      model_uri=META.local_name_source, domain=LocalName, range=Union[str, LocalNameLocalNameSource])

slots.local_name_value = Slot(uri=SKOS.altLabel, name="local_name_value", curie=SKOS.curie('altLabel'),
                      model_uri=META.local_name_value, domain=LocalName, range=str)

slots.local_names = Slot(uri=META.local_names, name="local_names", curie=META.curie('local_names'),
                      model_uri=META.local_names, domain=Element, range=Union[dict, "LocalName"])

slots.slot_definition_is_a = Slot(uri=META.is_a, name="slot_definition_is_a", curie=META.curie('is_a'),
                      model_uri=META.slot_definition_is_a, domain=SlotDefinition, range=Optional[Union[str, SlotDefinitionName]])

slots.slot_definition_mixins = Slot(uri=META.mixins, name="slot_definition_mixins", curie=META.curie('mixins'),
                      model_uri=META.slot_definition_mixins, domain=SlotDefinition, range=List[Union[str, SlotDefinitionName]])

slots.slot_definition_apply_to = Slot(uri=META.apply_to, name="slot_definition_apply_to", curie=META.curie('apply_to'),
                      model_uri=META.slot_definition_apply_to, domain=SlotDefinition, range=List[Union[str, SlotDefinitionName]])

slots.class_definition_is_a = Slot(uri=META.is_a, name="class_definition_is_a", curie=META.curie('is_a'),
                      model_uri=META.class_definition_is_a, domain=ClassDefinition, range=Optional[Union[str, ClassDefinitionName]])

slots.class_definition_mixins = Slot(uri=META.mixins, name="class_definition_mixins", curie=META.curie('mixins'),
                      model_uri=META.class_definition_mixins, domain=ClassDefinition, range=List[Union[str, ClassDefinitionName]])

slots.class_definition_apply_to = Slot(uri=META.apply_to, name="class_definition_apply_to", curie=META.curie('apply_to'),
                      model_uri=META.class_definition_apply_to, domain=ClassDefinition, range=List[Union[str, ClassDefinitionName]])



import os
import re
import shutil
import sys
import textwrap
import unittest
from contextlib import redirect_stdout
from io import StringIO
from typing import Union, List, Optional, Callable

# Make sure you import click from here rather than the root
import click

from biolinkml import MODULE_DIR
from tests import sourcedir
from tests.utils.compare_directories import are_dir_trees_equal
from tests.utils.dirutils import make_and_clear_directory
# Stop click from doing a sys.exit
from tests.utils.rdf_comparator import compare_rdf


class CLIExitException(Exception):
    ...


def no_click_exit(_self, _code=0):
    raise CLIExitException()


click.core.Context.exit = no_click_exit


class ClickTestCase(unittest.TestCase):
    # The four variables below must be set by the inheriting class
    testdir: str = None
    click_ep = None
    prog_name: str = None

    soft_compare: Optional[str] = None
    test_base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))
    temp_base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'temp'))

    @classmethod
    def setUpClass(cls):
        assert cls.testdir is not None, "Test directory must be defined"

        cls.testdir_path = os.path.join(cls.test_base_dir, cls.testdir)
        cls.tmpdir_path = os.path.join(cls.temp_base_dir, cls.testdir)
        if not os.path.exists(cls.testdir_path):
            print(f"Creating {cls.testdir_path}\n")
            make_and_clear_directory(cls.testdir_path)
        cls.creation_messages = []
        cls.keep_temp_directory = False

    @classmethod
    def tearDownClass(cls):
        if not cls.keep_temp_directory and os.path.exists(cls.tmpdir_path):
            make_and_clear_directory(cls.tmpdir_path)
            cls.keep_temp_directory = False
        if cls.creation_messages:
            for msg in cls.creation_messages:
                print(msg, file=sys.stderr)
            cls.creation_messages = []
            assert False, "Tests failed because baseline files were being created"

    def n3_comparator(self, old_data: str, new_data: str) -> str:
        return compare_rdf(old_data, new_data, "n3")

    def rdf_comparator(self, old_data: str, new_data: str) -> str:
        return compare_rdf(old_data, new_data)

    def always_pass_comparator(self, old_data: str, new_data: str) -> str:
        """
        No-op comparator -- everyone passes!
        :param old_data:
        :param new_data:
        :return:
        """
        return None


    @staticmethod
    def closein_comparison(old_txt: str, new_txt: str) -> None:
        """ Assist with testing comparison -- zero in on the first difference in a big string

        @param old_txt:
        @param new_txt:
        """
        window = 30
        view = 120

        nw = nt = new_txt.strip()
        ow = ot = old_txt.strip()
        if ot != nt:
            offset = 0
            while nt and ot and nt[:window] == ot[:window]:
                offset += window
                nt = nt[window:]
                ot = ot[window:]
            offset = max(offset-view, 0)
            print("   - - EXPECTED - -")
            print(ow[offset:offset+view+view])
            print("\n   - - ACTUAL - -")
            print(nw[offset:offset+view+view])

    def do_test(self, args: Union[str, List[str]], testfile: Optional[str] = "", *, error: type(Exception) = None,
                filtr: Optional[Callable[[str], str]] = None, tox_wrap_fix: bool = False,
                bypass_soft_compare: bool = False, dirbase: Optional[str] = None,
                comparator: Callable[[type(unittest.TestCase), str, str, str], str] = None,) -> None:
        """ Execute a command test

        @param args: Argument string or list to command
        @param testfile: name of file to record output in.  If absent, using directory mode
        @param error: If present, we expect this error
        @param filtr: Filter to remove date and app specific information from text
        @param tox_wrap_fix: tox seems to wrap redirected output at 60 columns.  If true, try wrapping the test
        file before failing
        @param bypass_soft_compare: True means not to use the class level soft compare
        @param dirbase: If present, compare tmpdir_path (new) to testdir_path(old)
        @param comparator: If present, use this method for comparison
        """
        new_directory = os.path.join(self.tmpdir_path, dirbase) if dirbase else None

        outf = StringIO()
        arg_list = args.split() if isinstance(args, str) else args
        from tests.utils.generator_utils import BIOLINK_IMPORT_MAP_FNAME
        if arg_list and arg_list[0] != '--help':
            arg_list += ["--importmap", BIOLINK_IMPORT_MAP_FNAME]
        if error:
            with self.assertRaises(error):
                self.click_ep(arg_list, standalone_mode=False)
            return

        arg_list += ["--log_level", "INFO"]

        with redirect_stdout(outf):
            try:
                self.click_ep(arg_list, prog_name=self.prog_name, standalone_mode=False)
            except CLIExitException:
                pass

        if not testfile:
            if dirbase:
                old_directory = os.path.join(self.testdir_path, dirbase)
                if os.path.exists(old_directory):
                    diffs = are_dir_trees_equal(old_directory, new_directory)
                    if diffs:
                        self.__class__.keep_temp_directory = True
                        print(diffs)
                        self.assertTrue(False)
                else:
                    shutil.copytree(new_directory, old_directory)
                    self.__class__.creation_messages.append(f"Directory: {old_directory} created")
            else:
                print("Directory comparison needs to be added", file=sys.stderr)
        else:
            testfile_path = os.path.join(self.testdir_path, testfile)
            new_txt = outf.getvalue().replace('\r\n', '\n').strip()
            if filtr:
                new_txt = filtr(new_txt)
            if not os.path.exists(testfile_path):
                with open(testfile_path, 'w') as f:
                    f.write(outf.getvalue())
                self.__class__.creation_messages.append(f"File: {testfile_path} created")
            else:
                with open(testfile_path) as f:
                    old_txt = f.read().replace('\r\n', '\n').strip()
                if filtr:
                    old_txt = filtr(old_txt)

                if old_txt != new_txt and not comparator and tox_wrap_fix:
                    old_txt = textwrap.fill(old_txt, 60)
                    new_txt = textwrap.fill(new_txt, 60)
                if comparator:
                    compare_text = comparator(self, old_txt, new_txt)
                else:
                    compare_text = None if old_txt == new_txt else ''

                # If necessary, update the test file
                if compare_text is not None:
                    hint = "(Not an error) " if self.soft_compare else ""
                    print(f"\n***** {hint} Mismatch on: {os.path.relpath(testfile_path, MODULE_DIR)}. "
                          f"Remove file to refresh *****\n")
                    if self.soft_compare and not bypass_soft_compare:
                        print(f"{self.id()}: {self.soft_compare}")
                    elif not comparator:
                        if len(old_txt) > 10000:
                            print(self.closein_comparison(old_txt, new_txt))
                        self.maxDiff = None
                        self.assertEqual(old_txt, new_txt)
                    else:
                        print(compare_text)
                        self.assertFalse(True, "Mismatch")

    def temp_directory(self, base: str) -> str:
        """
        Create a temporary directory and return the path

        :param base:
        :return: directory
        """
        if not os.path.exists(self.tmpdir_path):
            make_and_clear_directory(self.tmpdir_path)
        new_directory = os.path.join(self.tmpdir_path, base)
        make_and_clear_directory(new_directory)
        return new_directory


if __name__ == '__main__':
    unittest.main()



import os
from typing import TextIO, Union

from biolinkml.generators import PYTHON_GEN_VERSION
from biolinkml.generators.pythongen import PythonGenerator
from biolinkml.meta import SchemaDefinition
from biolinkml.utils.formatutils import split_line, be


class NamespaceGenerator(PythonGenerator):
    generatorname = os.path.basename(__file__)
    generatorversion = PYTHON_GEN_VERSION
    valid_formats = ['py']
    visit_all_class_slots = False

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], format: str = valid_formats[0],
                 emit_metadata: bool = True, **kwargs):
        super().__init__(schema, format, emit_metadata, **kwargs)

    def gen_namespaces(self) -> str:
        return '\n\t\t'.join([
            f"CurieNamespace('{pfx.replace('.', '_')}', '{self.namespaces[pfx]}'),"
            for pfx in sorted(self.emit_prefixes) if pfx in self.namespaces
        ])

    def gen_schema(self) -> str:
        split_descripton = '\n#              '.join(split_line(be(self.schema.description), split_len=100))
        head = f'''# Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
# Generation date: {self.schema.generation_date}
# Schema: {self.schema.name}
#''' if self.emit_metadata else ''

        return f'''{head}
# id: {self.schema.id}
# description: {split_descripton}
# license: {be(self.schema.license)}

from collections import defaultdict
from typing import Iterable, Dict, Tuple

from biolinkml.utils.curienamespace import CurieNamespace

GENE = 'gene'
DISEASE = 'disease'
CHEMICAL_SUBSTANCE = 'chemical substance'

SYMBOL = 'Approved_Symbol'


class IdentifierResolverException(RuntimeError):
    pass


class BiolinkNameSpace:
    """
    Map of BioLink Model registered URI Namespaces
    """

    _namespaces = [
        {self.gen_namespaces()}
    ]

    # class level dictionaries

    _prefix_map: Dict[str, CurieNamespace] = {{}}

    @classmethod
    def _get_prefix_map(cls):
        if not cls._prefix_map:
            for ns in cls._namespaces:
                # index by upper case for uniformity of search
                cls._prefix_map[ns.prefix.upper()] = ns
        return cls._prefix_map

    @classmethod
    def parse_curie(cls, curie: str) -> Tuple[CurieNamespace, str]:
        """
        Parse a candidate CURIE
        :param curie: candidate curie string
        :return: CURIE namespace and object_id
        """
        found = CurieNamespace("", ""), curie  # default value if not a CURIE or unknown XMLNS prefix
        if ':' in curie:
            part = curie.split(":")
            # Normalize retrieval with upper case of prefix for lookup
            prefix = part[0].upper()
            if prefix in cls._get_prefix_map():
                found = cls._prefix_map[prefix], part[1]
        return found

    @classmethod
    def parse_uri(cls, uri: str) -> Tuple[CurieNamespace,  str]:
        """
        Parse a candidate URI
        :param uri: candidate URI string
        :return: namespace and object_id
        """
        found = CurieNamespace("", ""), uri   # default value returned if unknown URI namespace

        # TODO: is there a more efficient lookup scheme here than a linear search of namespaces?
        for ns in cls._namespaces:
            base_uri = str(ns)
            if uri.startswith(base_uri):
                # simple minded deletion of base_uri to give the object_id
                object_id = uri.replace(base_uri, "")
                found = ns, object_id
                break
        return found

    @classmethod
    def parse_identifier(cls,  identifier: str) -> Tuple[CurieNamespace,  str]:

        # trivial case of a null identifier?
        if not identifier:
            return CurieNamespace("", ""), ""

        # check if this is a candidate URI...
        if identifier.lower().startswith("http"):
            # guess that perhaps it is, so try to parse it
            return cls.parse_uri(identifier)

        else:  # attempt to parse as a CURIE
            return cls.parse_curie(identifier)


def object_id(identifier, keep_version=False) -> str:
    """
    Returns the core object_id of a CURIE, with or without the version suffix.
    Note:  not designed to be used with a URI (will give an invalid outcome)
    :param identifier: candidate CURIE identifier for processing
    :param keep_version: True if the version string suffix is to be retained in the identifier
    :return:
    """
    # trivial case: null input value?
    if not identifier:
        return identifier

    if ':' in identifier:
        identifier = identifier.split(":")[1]

    if not keep_version and '.' in identifier:
        identifier = identifier.split(".")[0]

    return identifier


def fix_curies(identifiers, prefix=''):
    """
    Applies the specified XMLNS prefix to (an) identifier(s) known
    to be "raw" IDs as keys in a dictionary or elements in a list (or a simple string)
    :param identifiers:
    :param prefix:
    :return:
    """
    if not prefix:
        # return identifiers without modification
        # Caller may already consider them in curie format
        return identifiers

    if isinstance(identifiers, dict):
        curie_dict = defaultdict(dict)
        for key in identifiers.keys():
            curie_dict[prefix + ':' + object_id(key, keep_version=True)] = identifiers[key]
        return curie_dict

    # identifiers assumed to be just a single object identifier
    elif isinstance(identifiers, str):
        # single string to convert
        return prefix + ':' + object_id(identifiers, keep_version=True)

    elif isinstance(identifiers, Iterable):
        return [prefix + ':' + object_id(x, keep_version=True) for x in identifiers]

    else:
        raise RuntimeError("fix_curie() is not sure how to fix an instance of data type '", type(identifiers))


def curie(identifier) -> str:
    # Ignore enpty strings
    if not identifier:
        return ""
    else:
        namespace: CurieNamespace
        identifier_object_id: str
        namespace, identifier_object_id = BiolinkNameSpace.parse_identifier(identifier)
        return namespace.curie(identifier_object_id)
'''
        





# id: http://example.org/tests/timepoint
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
from biolinkml.utils.metamodelcore import Bool, ElementIdentifier, NCName, NodeIdentifier, URI, URIorCURIE, XSDDate, XSDDateTime, XSDTime

metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
SHEX = CurieNamespace('shex', 'http://www.w3.org/ns/shex#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = CurieNamespace('', 'http://example.org/tests/timepoint/')


# Types
class TimeType(Time):
    """ A time object represents a (local) time of day, independent of any particular day """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "time type"
    type_model_uri = URIRef("http://example.org/tests/timepoint/TimeType")


class String(str):
    """ A character string """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = URIRef("http://example.org/tests/timepoint/String")


class Integer(int):
    """ An integer """
    type_class_uri = XSD.integer
    type_class_curie = "xsd:integer"
    type_name = "integer"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Integer")


class Boolean(Bool):
    """ A binary (true or false) value """
    type_class_uri = XSD.boolean
    type_class_curie = "xsd:boolean"
    type_name = "boolean"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Boolean")


class Float(float):
    """ A real number that conforms to the xsd:float specification """
    type_class_uri = XSD.float
    type_class_curie = "xsd:float"
    type_name = "float"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Float")


class Double(float):
    """ A real number that conforms to the xsd:double specification """
    type_class_uri = XSD.double
    type_class_curie = "xsd:double"
    type_name = "double"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Double")


class Time(XSDTime):
    """ A time object represents a (local) time of day, independent of any particular day """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "time"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Time")


class Date(XSDDate):
    """ a date (year, month and day) in an idealized calendar """
    type_class_uri = XSD.date
    type_class_curie = "xsd:date"
    type_name = "date"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Date")


class Datetime(XSDDateTime):
    """ The combination of a date and time """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "datetime"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Datetime")


class Uriorcurie(URIorCURIE):
    """ a URI or a CURIE """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uriorcurie"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Uriorcurie")


class Uri(URI):
    """ a complete URI """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uri"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Uri")


class Ncname(NCName):
    """ Prefix part of CURIE """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "ncname"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Ncname")


class Objectidentifier(ElementIdentifier):
    """ A URI or CURIE that represents an object in the model. """
    type_class_uri = SHEX.iri
    type_class_curie = "shex:iri"
    type_name = "objectidentifier"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Objectidentifier")


class Nodeidentifier(NodeIdentifier):
    """ A URI, CURIE or BNODE that represents a node in a model. """
    type_class_uri = SHEX.nonLiteral
    type_class_curie = "shex:nonLiteral"
    type_name = "nodeidentifier"
    type_model_uri = URIRef("http://example.org/tests/timepoint/Nodeidentifier")


# Class references
class GeographicLocationK(extended_str):
    pass


class GeographicLocationAtTimeK(GeographicLocationK):
    pass


@dataclass
class GeographicLocation(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef("http://example.org/tests/timepoint/GeographicLocation")
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "geographic location"
    class_model_uri: ClassVar[URIRef] = URIRef("http://example.org/tests/timepoint/GeographicLocation")

    k: Union[str, GeographicLocationK]

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.k is None:
            raise ValueError(f"k must be supplied")
        if not isinstance(self.k, GeographicLocationK):
            self.k = GeographicLocationK(self.k)
        super().__post_init__(**kwargs)


@dataclass
class GeographicLocationAtTime(GeographicLocation):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef("http://example.org/tests/timepoint/GeographicLocationAtTime")
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "geographic location at time"
    class_model_uri: ClassVar[URIRef] = URIRef("http://example.org/tests/timepoint/GeographicLocationAtTime")

    k: Union[str, GeographicLocationAtTimeK] = None
    timepoint: Optional[Union[str, TimeType]] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.k is None:
            raise ValueError(f"k must be supplied")
        if not isinstance(self.k, GeographicLocationAtTimeK):
            self.k = GeographicLocationAtTimeK(self.k)
        if self.timepoint is not None and not isinstance(self.timepoint, TimeType):
            self.timepoint = TimeType(self.timepoint)
        super().__post_init__(**kwargs)



# Slots
class slots:
    pass

slots.k = Slot(uri=DEFAULT_.k, name="k", curie=DEFAULT_.curie('k'),
                      model_uri=DEFAULT_.k, domain=GeographicLocation, range=Union[str, GeographicLocationK])

slots.timepoint = Slot(uri=DEFAULT_.timepoint, name="timepoint", curie=DEFAULT_.curie('timepoint'),
                      model_uri=DEFAULT_.timepoint, domain=GeographicLocationAtTime, range=Optional[Union[str, TimeType]])


import os
import unittest
from typing import Union, List, Optional

from biolinkml import LOCAL_METAMODEL_YAML_FILE
from biolinkml.utils.schemaloader import SchemaLoader
from tests import sourcedir
from tests.test_utils import inputdir
from tests.test_utils.support.base import Base, update_all_files


class SchemaSynopsisTestCase(Base):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.loader: SchemaLoader = None

    def loadit(self, source: str) -> None:
        self.loader = SchemaLoader(source)
        self.loader.resolve()

    """ Tests for various parts of the schema synopsis file """
    def eval_synopsis(self, base_name: str, *,  is_sourcedir: bool=False, source: Optional[str]=None) -> None:
        fn = os.path.join(sourcedir if is_sourcedir else inputdir, base_name + '.yaml') if not source else source
        self.loadit(fn)
        errors = '\n'.join(self.loader.synopsis.errors())
        self.eval_output(errors, base_name + '.errs')

        synopsis = self.loader.synopsis.summary()
        self.eval_output(synopsis, base_name + '.synopsis')
        self.assertFalse(update_all_files, "Updating base files -- rerun")

    def assert_warning(self, text: str) -> bool:
        return f'* {text}' in self.loader.synopsis.summary()

    def assert_error(self, text: Union[str, List[str]]) -> None:
        if isinstance(text, str):
            text = [text]
        self.assertEqual(text, [e.strip() for e in self.loader.synopsis.errors()])

    def test_meta_synopsis(self):
        """ Raise a flag if the number of classes, slots, types or other elements change in the model  """
        self.eval_synopsis('meta', source=LOCAL_METAMODEL_YAML_FILE)

    def test_unitialized_domain(self):
        self.loadit(os.path.join(inputdir, 'synopsis1.yaml'))
        # Owners check no longer occurs
        # self.assert_error('Slot s1 has no owners')
        self.assert_warning('Unspecified domain: s1')

    def test_applyto(self):
        self.eval_synopsis('synopsis2')


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import click
from biolinkml import METAMODEL_NAMESPACE
from biolinkml.generators.jsonldcontextgen import ContextGenerator
from biolinkml.generators.jsonldgen import JSONLDGenerator
from biolinkml.generators.rdfgen import RDFGenerator
from biolinkml.generators.shexgen import cli, ShExGenerator
from pyshex import ShExEvaluator
from rdflib import Graph
from tests import source_yaml_path, DO_SHEX_VALIDATION
from tests.test_scripts.clicktestcase import ClickTestCase
from tests.utils.dirutils import make_and_clear_directory
from tests.utils.generator_utils import BIOLINK_IMPORT_MAP


class GenShExTestCase(ClickTestCase):
    testdir = "genshex"
    click_ep = cli
    prog_name = "gen-shex"

    def test_help(self):
        self.do_test("--help", 'help')

    def test_meta(self):
        """ Generate various forms of the metamodel in ShEx """
        self.maxDiff = None
        self.do_test(source_yaml_path, 'metashex.shex')
        self.do_test(source_yaml_path + ' -f json', 'metashex.json')
        self.do_test(source_yaml_path + ' -f rdf', 'metashex.ttl')
        self.do_test(source_yaml_path + ' -f shex', 'metashex.shex')
        self.do_test(source_yaml_path + ' --metauris', 'metashexn.shex')
        self.do_test(source_yaml_path + ' -f json', 'metashex.json')
        self.do_test(source_yaml_path + ' -f rdf', 'metashex.ttl')
        self.do_test(source_yaml_path + ' -f shex', 'metashex.shex')
        self.do_test(source_yaml_path + f' -f xsv', 'meta_error', error=click.exceptions.BadParameter)

    def test_rdf_shex(self):
        """ Generate ShEx and RDF for the model and verify that the RDF represents a valid instance """
        test_dir = os.path.join(self.tmpdir_path, 'meta_conformance_test')
        make_and_clear_directory(self.tmpdir_path)
        make_and_clear_directory(test_dir)

        json_file = os.path.join(test_dir, 'meta.jsonld')
        json_str = JSONLDGenerator(source_yaml_path, importmap=BIOLINK_IMPORT_MAP).serialize()
        with open(json_file, 'w') as f:
            f.write(json_str)

        context_file = os.path.join(test_dir, 'metacontext.jsonld')
        ContextGenerator(source_yaml_path, importmap=BIOLINK_IMPORT_MAP).serialize(output=context_file)
        self.assertTrue(os.path.exists(context_file))

        rdf_file = os.path.join(test_dir, 'meta.ttl')
        RDFGenerator(source_yaml_path, importmap=BIOLINK_IMPORT_MAP).serialize(output=rdf_file, context=context_file)
        self.assertTrue(os.path.exists(rdf_file))

        shex_file = os.path.join(test_dir, 'meta.shex')
        ShExGenerator(source_yaml_path, importmap=BIOLINK_IMPORT_MAP).serialize(output=shex_file, collections=False)
        self.assertTrue(os.path.exists(shex_file))

        if DO_SHEX_VALIDATION:
            g = Graph()
            g.load(rdf_file, format='ttl')
            focus = METAMODEL_NAMESPACE.metamodel
            start = METAMODEL_NAMESPACE.SchemaDefinition
            results = ShExEvaluator(g, shex_file, focus, start).evaluate(debug=False)
            success = all(r.result for r in results)
            if not success:
                for r in results:
                    if not r.result:
                        print(r.reason)
            else:
                make_and_clear_directory(test_dir)
            self.assertTrue(success)
        else:
            print("*** ShEX validation step was skipped. Set: tests.__init__.DO_SHEX_VALIDATION to run it")

if __name__ == '__main__':
    unittest.main()



import unittest

from rdflib import Graph


class Issue103TestCase(unittest.TestCase):

    @unittest.skipIf(True, "JSON-LD 1.1 Prefix Issue is still not resolved")
    def test_jsonld_prefix(self):
        test_json = '''
        {
            "@context": {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "@vocab": "http://example.org"
            },
            "CHEBI:33709": {
                "rdf:label": "Amino Acid"
            }
        }
        '''

        g = Graph().parse(data=test_json, format="json-ld", prefix=True)
        rdfstr = g.serialize(format="turtle").decode()
        assert '@prefix CHEBI: <http://purl.obolibrary.org/obo/CHEBI_>' in rdfstr

        g = Graph().parse(data=test_json, format="json-ld", prefix=False)
        rdfstr = g.serialize(format="turtle").decode()
        assert '@prefix CHEBI: <http://purl.obolibrary.org/obo/CHEBI_>' not in rdfstr


if __name__ == '__main__':
    unittest.main()



import unittest

from rdflib import URIRef
from rdflib.namespace import SKOS

from biolinkml.utils.namespaces import Namespaces


class NamespacesTestCase(unittest.TestCase):
    def test_namespaces(self):
        ns = Namespaces()
        ns['meta'] = "https://w3id.org/biolink/metamodel/"
        ns.skos = SKOS
        self.assertEqual(str(ns.skos), str(SKOS))
        self.assertEqual(ns.skos.note, SKOS.note)
        ns.OIO = URIRef("http://www.geneontology.org/formats/oboInOwl")
        ns['dc'] = "http://example.org/dc/"         # Overrides 'dc' in semweb_context
        ns['l1'] = "http://example.org/subset/"
        ns['l2'] = "http://example.org/subset/test/"
        ns['l3'] = "http://example.org/subset/t"
        # This is now a warning instead of a value error
        # with self.assertRaises(ValueError):
        #     ns['OIO'] = URIRef("http://www.geneontology.org/formats/another")
        # try:
        #     ns.OIO = URIRef("http://www.geneontology.org/formats/another")
        # except ValueError as e:
        #     self.assertEqual("Namespace OIO is already mapped to http://www.geneontology.org/formats/oboInOwl", str(e))
        with self.assertRaises(ValueError):
            ns["123"] = "http://example.org/foo/"

        with self.assertRaises(KeyError):
            ns.FOO

        ns._default = ns['meta']
        ns._default = ns['meta']
        with self.assertRaises(ValueError):
            ns._default = "http://example.org/wrong/"
        del ns._default
        with self.assertRaises(KeyError):
            del ns._default
        self.assertIsNone(ns._default)
        ns._default = ns['meta']

        ns._base = "http://example.org/base/"
        ns._base = "http://example.org/base/"
        with self.assertRaises(ValueError):
            ns._base = "http://example.org/wrong/"
        del ns._base
        with self.assertRaises(KeyError):
            del ns._base
        ns._base = "http://example.org/wrong/"
        del ns._default
        ns.add_prefixmap('semweb_context')
        ns.add_prefixmap('monarch_context')
        self.assertEqual('https://monarchinitiative.org/', str(ns._default))
        del ns._default
        ns._default = ns['meta']
        self.assertEqual('l1:foo', ns.curie_for("http://example.org/subset/foo"))
        self.assertEqual('l2:foo', ns.curie_for("http://example.org/subset/test/foo"))
        self.assertEqual('l3:able/foo', ns.curie_for("http://example.org/subset/table/foo"))
        #no comment in skos?
        #self.assertEqual(SKOS.comment, ns.uri_for("skos:comment"))
        self.assertEqual(URIRef('http://example.org/dc/table'), ns.uri_for("dc:table"))
        self.assertEqual(ns.uri_for("http://something.org"), URIRef("http://something.org"))
        self.assertEqual('https://w3id.org/biolink/metamodel/Schema', str(ns.uri_for(":Schema")))
        self.assertEqual(URIRef('http://example.org/wrong/Base'), ns.uri_for("Base"))
        del ns._base
        with self.assertRaises(ValueError):
            ns.uri_for("Base")
        try:
            ns.uri_for("Base")
        except ValueError as e:
            self.assertEqual('Unknown CURIE prefix: @base', str(e))

        self.assertIsNone(ns.curie_for("http://google.com/test"))
        with self.assertRaises(ValueError):
            ns.uri_for("1abc:junk")


if __name__ == '__main__':
    unittest.main()



import os
import re
import unittest

from jsonasobj import JsonObj, loads, as_json

from biolinkml import METAMODEL_CONTEXT_URI, META_BASE_URI
from biolinkml.utils.context_utils import merge_contexts

json_1 = '{ "ex": "http://example.org/test/", "ex2": "http://example.org/test2/" }'
json_2 = '{ "foo": 17, "@context": { "ex": "http://example.org/test3/", "ex2": {"@id": "http://example.org/test4/" }}}'

context_output = """{
   "@context": [
      "file://local.jsonld",
      "https://w3id.org/biolink/biolinkml/context.jsonld",
      {
         "ex": "http://example.org/test/",
         "ex2": "http://example.org/test2/"
      },
      {
         "ex": "http://example.org/test3/",
         "ex2": {
            "@id": "http://example.org/test4/"
         }
      }
   ]
}"""


class ContextUtilsTestCase(unittest.TestCase):
    def test_merge_contexts(self):
        self.assertIsNone(merge_contexts())
        self.assertEqual('file://local.jsonld', merge_contexts("local.jsonld")['@context'])
        self.assertEqual('file://local.jsonld', merge_contexts(["local.jsonld"])['@context'])
        self.assertEqual(METAMODEL_CONTEXT_URI, merge_contexts(METAMODEL_CONTEXT_URI)['@context'])
        self.assertEqual(METAMODEL_CONTEXT_URI, merge_contexts([METAMODEL_CONTEXT_URI])['@context'])
        self.assertEqual(JsonObj(ex='http://example.org/test/', ex2='http://example.org/test2/'),
                         merge_contexts(json_1)['@context'])
        self.assertEqual(JsonObj(ex='http://example.org/test/', ex2='http://example.org/test2/'),
                         merge_contexts([json_1])['@context'])
        self.assertEqual(JsonObj(ex='http://example.org/test3/', ex2=JsonObj(**{'@id': 'http://example.org/test4/'})),
                         merge_contexts(json_2)['@context'])
        self.assertEqual(JsonObj(ex='http://example.org/test3/', ex2=JsonObj(**{'@id': 'http://example.org/test4/'})),
                         merge_contexts([json_2])['@context'])
        self.assertEqual([f'file://local.jsonld',
                          'https://w3id.org/biolink/biolinkml/context.jsonld',
                          JsonObj(ex='http://example.org/test/', ex2='http://example.org/test2/'),
                          JsonObj(ex='http://example.org/test3/', ex2=JsonObj(**{'@id': 'http://example.org/test4/'}))],
                         merge_contexts(["local.jsonld", METAMODEL_CONTEXT_URI, json_1, json_2])['@context'])
        self.assertEqual(loads(context_output),
                         merge_contexts(["local.jsonld", METAMODEL_CONTEXT_URI, json_1, json_2]))

        # Dups are not removed
        self.assertEqual(
            JsonObj(**{'@context': [JsonObj(ex='http://example.org/test/', ex2='http://example.org/test2/'),
                                    JsonObj(ex='http://example.org/test/', ex2='http://example.org/test2/')]}),
            merge_contexts([json_1, json_1]))
        self.assertEqual('file://local.jsonld', merge_contexts("local.jsonld")['@context'])

    def test_merge_contexts_base(self):
        self.assertEqual(
            JsonObj(**{'@context':
                           JsonObj(**{'@base': 'file://relloc'})}),
            merge_contexts(base='file://relloc'))
        self.assertEqual(loads(f'{{"@context": {{"@base": "{META_BASE_URI}"}}}}'), merge_contexts(base=META_BASE_URI))
        self.assertEqual(loads("""
{"@context": [
      "https://w3id.org/biolink/biolinkml/context.jsonld",
      {
         "ex": "http://example.org/test/",
         "ex2": "http://example.org/test2/"
      },
      {
         "ex": "http://example.org/test3/",
         "ex2": {
            "@id": "http://example.org/test4/"
         }
      },
      {
         "@base": "https://w3id.org/biolink/biolinkml/"
      }
   ]
}"""), merge_contexts([METAMODEL_CONTEXT_URI, json_1, json_2], base=META_BASE_URI))


if __name__ == '__main__':
    unittest.main()




""" Metadata filters for test cases -- various tools to remove metadata from output """
import re


def ldcontext_metadata_filter(s: str) -> str:
    """
    Metafilter for jsonld context
    :param s: jsonld context string
    :return: string stripped of metadata
    """
    return re.sub(r'Auto generated from .*? by', 'Auto generated from ... by',
                  re.sub(r'Generation date: .*?\\n', r'Generation date: \\n', s))


def json_metadata_filter(s: str) -> str:
    return re.sub(r'("source_file_date": )".*"', r'\1"Friday Sep 27 12:00:00 2003"',
           re.sub(r'("generation_date": )".*"', r'\1"Friday Sep 27 12:00:00 2003"',
           re.sub(r'("source_file_size": )".*"', r'\1"23600:', s)))


def json_metadata_context_filter(s: str) -> str:
    return re.sub(r'file:///.*/', r'', json_metadata_filter(s))


def metadata_filter(s: str) -> str:
    return re.sub(r'(# Auto generated from ).*(\.yaml by pythongen\.py version:) .*', r'\1\2',
                  re.sub(r'(# Generation date:) .*', r'\1', re.sub(r'\r\n', '\n', s)))